# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

u"""Provide time related exceptions and functions"""
from __future__ import division

from past.utils import old_div
from builtins import map

import time
import types
import re
import calendar
import sys
from duplicity import config
from duplicity import util

# For type testing against both int and long types that works in python 2/3
if sys.version_info < (3,):
    integer_types = (int, types.LongType)
else:
    integer_types = (int,)


class TimeException(Exception):
    pass


_interval_conv_dict = {u"s": 1, u"m": 60, u"h": 3600, u"D": 86400,
                       u"W": 7 * 86400, u"M": 30 * 86400, u"Y": 365 * 86400}
_integer_regexp = re.compile(u"^[0-9]+$")
_interval_regexp = re.compile(u"^([0-9]+)([smhDWMY])")
_genstr_date_regexp1 = re.compile(u"^(?P<year>[0-9]{4})[-/]"
                                  u"(?P<month>[0-9]{1,2})[-/]"
                                  u"(?P<day>[0-9]{1,2})$")
_genstr_date_regexp2 = re.compile(u"^(?P<month>[0-9]{1,2})[-/]"
                                  u"(?P<day>[0-9]{1,2})[-/]"
                                  u"(?P<year>[0-9]{4})$")
_genstr_date_regexp3 = re.compile(u"^(?P<year>[0-9]{4})"
                                  u"(?P<month>[0-9]{2})"
                                  u"(?P<day>[0-9]{2})Z$")
curtime = curtimestr = None
prevtime = prevtimestr = None

bad_interval_string = _(u"""Bad interval string "%s"

Intervals are specified like 2Y (2 years) or 2h30m (2.5 hours).  The
allowed special characters are s, m, h, D, W, M, and Y.  See the man
page for more information.""")

bad_time_string = _(u"""Bad time string "%s"

The acceptible time strings are intervals (like "3D64s"), w3-datetime
strings, like "2002-04-26T04:22:01-07:00" (strings like
"2002-04-26T04:22:01" are also acceptable - duplicity will use the
current time zone), or ordinary dates like 2/4/1997 or 2001-04-23
(various combinations are acceptable, but the month always precedes
the day).""")


def setcurtime(time_in_secs=None):
    u"""Sets the current time in curtime and curtimestr"""
    global curtime, curtimestr
    t = time_in_secs or int(time.time())
    assert type(t) in integer_types
    curtime, curtimestr = t, timetostring(t)


def setprevtime(time_in_secs):
    u"""Sets the previous time in prevtime and prevtimestr"""
    global prevtime, prevtimestr
    assert type(time_in_secs) in integer_types, prevtime
    prevtime, prevtimestr = time_in_secs, timetostring(time_in_secs)


def timetostring(timeinseconds):
    u"""Return w3 or duplicity datetime compliant listing of timeinseconds"""

    if config.old_filenames:
        # We need to know if DST applies to append the correct offset. So
        #    1. Save the tuple returned by localtime.
        #    2. Pass the DST flag into gettzd
        lcltime = time.localtime(timeinseconds)
        return time.strftime(u"%Y-%m-%dT%H" + config.time_separator +
                             u"%M" + config.time_separator + u"%S",
                             lcltime) + gettzd(lcltime[-1])
    else:
        # DST never applies to UTC
        lcltime = time.gmtime(timeinseconds)
        return time.strftime(u"%Y%m%dT%H%M%SZ", lcltime)


def stringtotime(timestring):
    u"""Return time in seconds from w3 or duplicity timestring

    If there is an error parsing the string, or it doesn't look
    like a valid datetime string, return None.
    """
    try:
        date, daytime = timestring[:19].split(u"T")
        if len(timestring) == 16:
            # new format for filename time
            year, month, day = list(map(int,
                                    [date[0:4], date[4:6], date[6:8]]))
            hour, minute, second = list(map(int,
                                        [daytime[0:2], daytime[2:4], daytime[4:6]]))
        else:
            # old format for filename time
            year, month, day = list(map(int, date.split(u"-")))
            hour, minute, second = list(map(int,
                                        daytime.split(config.time_separator)))
        assert 1900 < year < 2100, year
        assert 1 <= month <= 12
        assert 1 <= day <= 31
        assert 0 <= hour <= 23
        assert 0 <= minute <= 59
        assert 0 <= second <= 61  # leap seconds
        # We want to return the time in units of seconds since the
        # epoch. Unfortunately the only functin that does this
        # works in terms of the current timezone and we have a
        # timezone offset in the string.
        timetuple = (year, month, day, hour, minute, second, -1, -1, 0)

        if len(timestring) == 16:
            # as said in documentation, time.gmtime() and timegm() are each others' inverse.
            # As far as UTC format is used in new file format,
            # do not rely on system's python DST and tzdata settings
            # and use functions that working with UTC
            utc_in_secs = calendar.timegm(timetuple)
        else:
            # mktime assumed that the tuple was a local time. Compensate
            # by subtracting the value for the current timezone.
            # We don't need to worry about DST here because we turned it
            # off in the tuple
            local_in_secs = time.mktime(timetuple)
            utc_in_secs = local_in_secs - time.timezone
        # Now apply the offset that we were given in the time string
        # This gives the correct number of seconds from the epoch
        # even when we're not in the same timezone that wrote the
        # string
        if len(timestring) == 16:
            return int(utc_in_secs)
        else:
            return int(utc_in_secs + tzdtoseconds(timestring[19:]))
    except (TypeError, ValueError, AssertionError):
        return None


def timetopretty(timeinseconds):
    u"""Return pretty version of time"""
    return time.asctime(time.localtime(timeinseconds))


def stringtopretty(timestring):
    u"""Return pretty version of time given w3 time string"""
    return timetopretty(stringtotime(timestring))


def inttopretty(seconds):
    u"""Convert num of seconds to readable string like "2 hours"."""
    partlist = []
    hours, seconds = divmod(seconds, 3600)
    if hours > 1:
        partlist.append(u"%d hours" % hours)
    elif hours == 1:
        partlist.append(u"1 hour")

    minutes, seconds = divmod(seconds, 60)
    if minutes > 1:
        partlist.append(u"%d minutes" % minutes)
    elif minutes == 1:
        partlist.append(u"1 minute")

    if seconds == 1:
        partlist.append(u"1 second")
    elif not partlist or seconds > 1:
        if isinstance(seconds, integer_types):
            partlist.append(u"%s seconds" % seconds)
        else:
            partlist.append(u"%.2f seconds" % seconds)
    return u" ".join(partlist)


def intstringtoseconds(interval_string):
    u"""Convert a string expressing an interval (e.g. "4D2s") to seconds"""
    def error():
        raise TimeException(bad_interval_string % util.escape(interval_string))

    if len(interval_string) < 2:
        error()

    total = 0
    while interval_string:
        match = _interval_regexp.match(interval_string)
        if not match:
            error()
        num, ext = int(match.group(1)), match.group(2)
        if ext not in _interval_conv_dict or num < 0:
            error()
        total += num * _interval_conv_dict[ext]
        interval_string = interval_string[match.end(0):]
    return total


def gettzd(dstflag):
    u"""Return w3's timezone identification string.

    Expresed as [+/-]hh:mm.  For instance, PST is -08:00.  Zone is
    coincides with what localtime(), etc., use.

    """
    # time.daylight doesn't help us. It's a flag that indicates that we
    # have a dst option for the current timezone. Compensate by allowing
    # the caller to pass a flag to indicate that DST applies. This flag
    # is in the same format as the last member of the tuple returned by
    # time.localtime()

    if dstflag > 0:
        offset = old_div(-1 * time.altzone, 60)
    else:
        offset = old_div(-1 * time.timezone, 60)
    if offset > 0:
        prefix = u"+"
    elif offset < 0:
        prefix = u"-"
    else:
        return u"Z"  # time is already in UTC

    hours, minutes = list(map(abs, divmod(offset, 60)))
    assert 0 <= hours <= 23
    assert 0 <= minutes <= 59
    return u"%s%02d%s%02d" % (prefix, hours, config.time_separator, minutes)


def tzdtoseconds(tzd):
    u"""Given w3 compliant TZD, return how far ahead UTC is"""
    if tzd == u"Z":
        return 0
    assert len(tzd) == 6  # only accept forms like +08:00 for now
    assert (tzd[0] == u"-" or tzd[0] == u"+") and \
        tzd[3] == config.time_separator
    return -60 * (60 * int(tzd[:3]) + int(tzd[4:]))


def cmp(time1, time2):
    u"""Compare time1 and time2 and return -1, 0, or 1"""
    if isinstance(time1, (str, u"".__class__)):
        time1 = stringtotime(time1)
        assert time1 is not None
    if isinstance(time2, (str, u"".__class__)):
        time2 = stringtotime(time2)
        assert time2 is not None

    if time1 < time2:
        return -1
    elif time1 == time2:
        return 0
    else:
        return 1


def genstrtotime(timestr, override_curtime=None):
    u"""Convert a generic time string to a time in seconds"""
    if override_curtime is None:
        override_curtime = curtime
    if timestr == u"now":
        return override_curtime

    def error():
        raise TimeException(bad_time_string % util.escape(timestr))

    # Test for straight integer
    if _integer_regexp.search(timestr):
        return int(timestr)

    # Test for w3-datetime format, possibly missing tzd
    # This is an ugly hack. We need to know if DST applies when doing
    # gettzd. However, we don't have the flag to pass. Assume that DST
    # doesn't apply and pass 0. Getting a reasonable default from
    # localtime() is a bad idea, since we transition to/from DST between
    # calls to this method on the same run

    t = stringtotime(timestr) or stringtotime(timestr + gettzd(0))
    if t:
        return t

    try:  # test for an interval, like "2 days ago"
        return override_curtime - intstringtoseconds(timestr)
    except TimeException:
        pass

    # Now check for dates like 2001/3/23
    match = (_genstr_date_regexp1.search(timestr) or
             _genstr_date_regexp2.search(timestr) or
             _genstr_date_regexp3.search(timestr))
    if not match:
        error()
    timestr = u"%s-%02d-%02dT00:00:00%s" % (match.group(u'year'),
                                            int(match.group(u'month')),
                                            int(match.group(u'day')),
                                            gettzd(0))
    t = stringtotime(timestr)
    if t:
        return t
    else:
        error()



# -*- coding: utf-8 -*-
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2014 Aaron Whitehouse <aaron@whitehouse.kiwi.nz>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()


import os
import sys
import platform
import io

import unittest

from . import FunctionalTestCase


class IncludeExcludeFunctionalTest(FunctionalTestCase):
    u"""
    This contains methods used in the tests below for testing the include, exclude and various filelist features.
    """

    # These tests assume the following files and logic, with:
    # "is" meaning that the file is included specifically
    # "ia" meaning that the file should be included automatically because its parent is included
    # "ic" meaning that the folder is included because its contents are included
    # "es" meaning that the file is excluded specifically
    # "ea" meaning that the file should be excluded automatically because its parent is included
    # select2 (es)
    # --- 1.doc (ea)
    # --- 1.py (is)
    # --- 1 (is)
    # ------ 1sub1 (ia)
    # --------- 1sub1sub1 (ia)
    # ------------ 1sub1sub1_file.txt (ia)
    # --------- 1sub1sub2 (es)
    # ------------ 1sub1sub2_file.txt (ea)
    # --------- 1sub1sub3 (ia)
    # ------------ 1sub1sub3_file.txt (es)
    # ------ 1sub2 (ic)
    # --------- 1sub2sub1 (is)
    # --------- 1sub2sub2 (ea)
    # --------- 1sub2sub3 (es)  # Not necessary as also ea, but to ensure there are no issues doing so
    # ------ 1sub3 (ia)
    # --------- 1sub3sub1 (es)
    # --------- 1sub3sub2 (es)
    # --------- 1sub3sub3 (ia)
    # --- 2 (ic)
    # ------ 2sub1 (is)
    # --------- 2sub1sub1 (ia)
    # ------------ 2sub1sub1_file.txt (ia)
    # --------- 2sub1sub2 (es)
    # --------- 2sub1sub3 (es)
    # ------ 2sub2 (ea)
    # --------- 2sub2sub1 (ea)
    # --------- 2sub2sub2 (ea)
    # --------- 2sub2sub3 (ea)
    # ------ 2sub3 (ea)
    # --------- 2sub3sub1 (ea)
    # --------- 2sub3sub3 (ea)
    # --------- 2sub3sub2 (ea)
    # --- 3 (is)
    # ------ 3sub1 (es)
    # --------- 3sub1sub1 (ea)
    # --------- 3sub1sub2 (ea)
    # --------- 3sub1sub3 (ea)
    # ------ 3sub2 (ia)
    # --------- 3sub2sub1 (ia)
    # --------- 3sub2sub2 (ia)
    # --------- 3sub2sub3 (ia)
    # ------ 3sub3 (is)  # Not necessary as also ia, but to ensure there are no issues doing so
    # --------- 3sub3sub1 (ia)
    # --------- 3sub3sub2 (es, ic)
    # ------------ 3sub3sub2_file.txt (is)
    # --------- 3sub3sub3 (ia)
    # --- trailing_space  (ea)  # Note this is "trailing_space ". Excluded until trailing_space test, when (is)
    # ------ trailing_space sub1 (ea)  # Excluded until trailing_space test, when (ia)
    # ------ trailing_space sub2 (ea)  # Excluded until trailing_space test, when (es, ic)
    # --------- trailing_space sub2_file.txt (ea)  # Excluded until trailing_space test, when (is)

    complete_directory_tree = [
        [u"1", u"2", u"3", u"trailing_space ", u"1.doc", u"1.py"],
        [u"1sub1", u"1sub2", u"1sub3"],
        [u"1sub1sub1", u"1sub1sub2", u"1sub1sub3"],
        [u"1sub1sub1_file.txt"],
        [u"1sub1sub2_file.txt"],
        [u"1sub1sub3_file.txt"],
        [u"1sub2sub1", u"1sub2sub2", u"1sub2sub3"],
        [u"1sub3sub1", u"1sub3sub2", u"1sub3sub3"],
        [u"2sub1", u"2sub2", u"2sub3"],
        [u"2sub1sub1", u"2sub1sub2", u"2sub1sub3"],
        [u"2sub1sub1_file.txt"],
        [u"2sub2sub1", u"2sub2sub2", u"2sub2sub3"],
        [u"2sub3sub1", u"2sub3sub2", u"2sub3sub3"],
        [u"3sub1", u"3sub2", u"3sub3"],
        [u"3sub1sub1", u"3sub1sub2", u"3sub1sub3"],
        [u"3sub2sub1", u"3sub2sub2", u"3sub2sub3"],
        [u"3sub3sub1", u"3sub3sub2", u"3sub3sub3"],
        [u"3sub3sub2_file.txt"],
        [u"trailing_space sub1", u"trailing_space sub2"],
        [u"trailing_space sub2_file.txt"]
    ]

    expected_restored_tree = [[u"1", u"2", u"3", u"1.py"],
                              [u"1sub1", u"1sub2", u"1sub3"],
                              [u"1sub1sub1", u"1sub1sub3"],
                              [u"1sub1sub1_file.txt"],
                              [u"1sub2sub1"],
                              [u"1sub3sub3"],
                              [u"2sub1"],
                              [u"2sub1sub1"],
                              [u"2sub1sub1_file.txt"],
                              [u"3sub2", u"3sub3"],
                              [u"3sub2sub1", u"3sub2sub2", u"3sub2sub3"],
                              [u"3sub3sub1", u"3sub3sub2", u"3sub3sub3"],
                              [u"3sub3sub2_file.txt"]]

    expected_restored_tree_with_trailing_space = [[u"1", u"2", u"3", u"trailing_space ", u"1.py"],
                                                  [u"1sub1", u"1sub2", u"1sub3"],
                                                  [u"1sub1sub1", u"1sub1sub3"],
                                                  [u"1sub1sub1_file.txt"],
                                                  [u"1sub2sub1"],
                                                  [u"1sub3sub3"],
                                                  [u"2sub1"],
                                                  [u"2sub1sub1"],
                                                  [u"2sub1sub1_file.txt"],
                                                  [u"3sub2", u"3sub3"],
                                                  [u"3sub2sub1", u"3sub2sub2", u"3sub2sub3"],
                                                  [u"3sub3sub1", u"3sub3sub2", u"3sub3sub3"],
                                                  [u"3sub3sub2_file.txt"],
                                                  [u"trailing_space sub1", u"trailing_space sub2"],
                                                  [u"trailing_space sub2_file.txt"]]

    def directory_tree_to_list_of_lists(self, parent_directory):
        u"""
        This takes a folder as an input and returns a list with its contents. If the directory has subdirectories, it
        returns a list of lists with the contents of those subdirectories.
        """
        directory_list = []
        for root, dirs, files in os.walk(parent_directory):
            to_add = []
            if dirs:
                dirs.sort()  # So that we can easily compare to what we expect
                to_add = dirs
            if files:
                files.sort()  # So that we can easily compare to what we expect
                to_add += files
            if to_add:
                directory_list.append(to_add)
        return directory_list


class TestCheckTestFiles(IncludeExcludeFunctionalTest):
    u""" Tests the testfiles required by the exclude/include tests are as expected. """

    def test_files_are_as_expected(self):
        u"""Test that the contents of testfiles/select are as expected."""
        testfiles = self.directory_tree_to_list_of_lists(u"testfiles/select2")
        # print(testfiles)
        self.assertEqual(testfiles, self.complete_directory_tree)


class TestIncludeExcludeOptions(IncludeExcludeFunctionalTest):
    u""" This tests the behaviour of the duplicity binary when the include/exclude options are passed directly """

    def test_include_exclude_basic(self):
        u""" Test --include and --exclude work in the basic case """
        self.backup(u"full", u"testfiles/select2",
                    options=[u"--include", u"testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt",
                             u"--exclude", u"testfiles/select2/3/3sub3/3sub3sub2",
                             u"--include", u"testfiles/select2/3/3sub2/3sub2sub2",
                             u"--include", u"testfiles/select2/3/3sub3",
                             u"--exclude", u"testfiles/select2/3/3sub1",
                             u"--exclude", u"testfiles/select2/2/2sub1/2sub1sub3",
                             u"--exclude", u"testfiles/select2/2/2sub1/2sub1sub2",
                             u"--include", u"testfiles/select2/2/2sub1",
                             u"--exclude", u"testfiles/select2/1/1sub3/1sub3sub2",
                             u"--exclude", u"testfiles/select2/1/1sub3/1sub3sub1",
                             u"--exclude", u"testfiles/select2/1/1sub2/1sub2sub3",
                             u"--include", u"testfiles/select2/1/1sub2/1sub2sub1",
                             u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt",
                             u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub2",
                             u"--exclude", u"testfiles/select2/1/1sub2",
                             u"--include", u"testfiles/select2/1.py",
                             u"--include", u"testfiles/select2/3",
                             u"--include", u"testfiles/select2/1",
                             u"--exclude", u"testfiles/select2/**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_include_exclude_trailing_whitespace(self):
        u"""Test that folders with trailing whitespace in the names work correctly when passing as include/exclude"""
        # Note that, because this only passes items in as a list of options, this test does not test whether duplicity
        # would correctly interpret commandline options with spaces. However, bin/duplicity uses sys.argv[1:], which
        # should return a list of strings after having correctly processed quotes etc.
        self.backup(u"full", u"testfiles/select2",
                    options=[u"--include",
                             u"testfiles/select2/trailing_space /trailing_space sub2/trailing_space sub2_file.txt",
                             u"--exclude", u"testfiles/select2/trailing_space /trailing_space sub2",
                             u"--include", u"testfiles/select2/trailing_space ",
                             u"--include", u"testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt",
                             u"--exclude", u"testfiles/select2/3/3sub3/3sub3sub2",
                             u"--include", u"testfiles/select2/3/3sub2/3sub2sub2",
                             u"--include", u"testfiles/select2/3/3sub3",
                             u"--exclude", u"testfiles/select2/3/3sub1",
                             u"--exclude", u"testfiles/select2/2/2sub1/2sub1sub3",
                             u"--exclude", u"testfiles/select2/2/2sub1/2sub1sub2",
                             u"--include", u"testfiles/select2/2/2sub1",
                             u"--exclude", u"testfiles/select2/1/1sub3/1sub3sub2",
                             u"--exclude", u"testfiles/select2/1/1sub3/1sub3sub1",
                             u"--exclude", u"testfiles/select2/1/1sub2/1sub2sub3",
                             u"--include", u"testfiles/select2/1/1sub2/1sub2sub1",
                             u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt",
                             u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub2",
                             u"--exclude", u"testfiles/select2/1/1sub2",
                             u"--include", u"testfiles/select2/1.py",
                             u"--include", u"testfiles/select2/3",
                             u"--include", u"testfiles/select2/1",
                             u"--exclude", u"testfiles/select2/**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree_with_trailing_space)


class TestExcludeFilelistTest(IncludeExcludeFunctionalTest):
    u"""
    Test --exclude-filelist using duplicity binary.
    """

    def test_exclude_filelist(self):
        u"""Test that exclude filelist works in the basic case """
        # As this is an exclude filelist any lines with no +/- modifier should be treated as if they have a -.
        # Create a filelist
        with io.open(u"testfiles/exclude.txt", u"w") as f:
            f.write(u"+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n"
                    u"testfiles/select2/3/3sub3/3sub3sub2\n"
                    u"+ testfiles/select2/3/3sub2/3sub2sub2\n"
                    u"+ testfiles/select2/3/3sub3\n"
                    u"- testfiles/select2/3/3sub1\n"  # - added to ensure it makes no difference
                    u"testfiles/select2/2/2sub1/2sub1sub3\n"
                    u"testfiles/select2/2/2sub1/2sub1sub2\n"
                    u"+ testfiles/select2/2/2sub1\n"
                    u"testfiles/select2/1/1sub3/1sub3sub2\n"
                    u"testfiles/select2/1/1sub3/1sub3sub1\n"
                    u"testfiles/select2/1/1sub2/1sub2sub3\n"
                    u"+ testfiles/select2/1/1sub2/1sub2sub1\n"
                    u"testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n"
                    u"testfiles/select2/1/1sub1/1sub1sub2\n"
                    u"- testfiles/select2/1/1sub2\n"  # - added to ensure it makes no difference
                    u"+ testfiles/select2/1.py\n"
                    u"+ testfiles/select2/3\n"
                    u"+ testfiles/select2/1\n"
                    u"testfiles/select2/**")
        self.backup(u"full", u"testfiles/select2", options=[u"--exclude-filelist=testfiles/exclude.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_exclude_filelist_combined_imperfections(self):
        u"""Test that exclude filelist works with imperfections in the input file"""
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier (both " and ')

        # Create a filelist
        with io.open(u"testfiles/exclude.txt", u"w") as f:
            f.write(u"+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n"
                    u"testfiles/select2/3/3sub3/3sub3sub2\n"
                    u"+ testfiles/select2/3/3sub2/3sub2sub2\n"
                    u" + testfiles/select2/3/3sub3\n"  # Note leading space added here
                    u"- testfiles/select2/3/3sub1\n"
                    u"  testfiles/select2/2/2sub1/2sub1sub3\n"  # Note leading spaces added here
                    u"\n"
                    u"testfiles/select2/2/2sub1/2sub1sub2\n"
                    u" + testfiles/select2/2/2sub1 \n"  # Note added trailing/leading space here
                    u'- "testfiles/select2/1/1sub3/1sub3sub2"\n'  # Unnecessary quotes
                    u"# Testing a full-line comment\n"
                    u"'testfiles/select2/1/1sub3/1sub3sub1'  \n"  # Note added spaces and quotes here
                    u"testfiles/select2/1/1sub2/1sub2sub3\n"
                    u"    \n"
                    u"+ testfiles/select2/1/1sub2/1sub2sub1\n"
                    u"- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n"
                    u"testfiles/select2/1/1sub1/1sub1sub2\n"
                    u"     # Testing a full-line comment with leading and trailing spaces     \n"
                    u"testfiles/select2/1/1sub2  \n"  # Note added spaces here
                    u"+ testfiles/select2/1.py\n"
                    u"+ testfiles/select2/3 \n"  # Note added space here
                    u"+ testfiles/select2/1\n"
                    u"- testfiles/select2/**")
        self.backup(u"full", u"testfiles/select2", options=[u"--exclude-filelist=testfiles/exclude.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_exclude_globbing_filelist_combined_imperfections(self):
        u"""Test that exclude globbing filelist works with imperfections in the input file"""
        # Identical to test_exclude_filelist_combined_imperfections and included to ensure that
        # the deprecated --exclude-globbing-filelist function works as expected until it is deliberately removed.
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier (both " and ')

        # Create a filelist
        with io.open(u"testfiles/exclude.txt", u"w") as f:
            f.write(u"+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n"
                    u"testfiles/select2/3/3sub3/3sub3sub2\n"
                    u"+ testfiles/select2/3/3sub2/3sub2sub2\n"
                    u" + testfiles/select2/3/3sub3\n"  # Note leading space added here
                    u"- testfiles/select2/3/3sub1\n"
                    u"  testfiles/select2/2/2sub1/2sub1sub3\n"  # Note leading spaces added here
                    u"\n"
                    u"testfiles/select2/2/2sub1/2sub1sub2\n"
                    u" + testfiles/select2/2/2sub1 \n"  # Note added trailing/leading space here
                    u'- "testfiles/select2/1/1sub3/1sub3sub2"\n'  # Unnecessary quotes
                    u"# Testing a full-line comment\n"
                    u"'testfiles/select2/1/1sub3/1sub3sub1'  \n"  # Note added spaces and quotes here
                    u"testfiles/select2/1/1sub2/1sub2sub3\n"
                    u"    \n"
                    u"+ testfiles/select2/1/1sub2/1sub2sub1\n"
                    u"- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n"
                    u"testfiles/select2/1/1sub1/1sub1sub2\n"
                    u"     # Testing a full-line comment with leading and trailing spaces     \n"
                    u"testfiles/select2/1/1sub2  \n"  # Note added spaces here
                    u"+ testfiles/select2/1.py\n"
                    u"+ testfiles/select2/3 \n"  # Note added space here
                    u"+ testfiles/select2/1\n"
                    u"- testfiles/select2/**")
        self.backup(u"full", u"testfiles/select2", options=[u"--exclude-globbing-filelist=testfiles/exclude.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_exclude_filelist_trailing_whitespace_folders_work_with_quotes(self):
        u"""Test that folders with trailing whitespace in the names work correctly if they are enclosed in quotes"""
        # Create a filelist
        with io.open(u"testfiles/exclude.txt", u"w") as f:
            f.write(u'+ "testfiles/select2/trailing_space /trailing_space sub2/trailing_space sub2_file.txt"\n'  # New
                    u"- 'testfiles/select2/trailing_space /trailing_space sub2'\n"  # New
                    u'+ "testfiles/select2/trailing_space "\n'  # New
                    u"+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n"
                    u"testfiles/select2/3/3sub3/3sub3sub2\n"
                    u"+ testfiles/select2/3/3sub2/3sub2sub2\n"
                    u"+ testfiles/select2/3/3sub3\n"
                    u"- testfiles/select2/3/3sub1\n"
                    u"testfiles/select2/2/2sub1/2sub1sub3\n"
                    u"testfiles/select2/2/2sub1/2sub1sub2\n"
                    u"+ testfiles/select2/2/2sub1\n"
                    u"testfiles/select2/1/1sub3/1sub3sub2\n"
                    u"testfiles/select2/1/1sub3/1sub3sub1\n"
                    u"testfiles/select2/1/1sub2/1sub2sub3\n"
                    u"+ testfiles/select2/1/1sub2/1sub2sub1\n"
                    u"testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n"
                    u"testfiles/select2/1/1sub1/1sub1sub2\n"
                    u"- testfiles/select2/1/1sub2\n"
                    u"+ testfiles/select2/1.py\n"
                    u"+ testfiles/select2/3\n"
                    u"+ testfiles/select2/1\n"
                    u"testfiles/select2/**")
        self.backup(u"full", u"testfiles/select2", options=[u"--exclude-filelist=testfiles/exclude.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree_with_trailing_space)

    def test_exclude_filelist_progress_option(self):
        u"""Test that exclude filelist is unaffected by the --progress option"""
        # Regression test for Bug #1264744 (https://bugs.launchpad.net/duplicity/+bug/1264744)
        # Create a filelist identical to that used in test_exclude_filelist
        with io.open(u"testfiles/exclude.txt", u"w") as f:
            f.write(u"+ testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n"
                    u"testfiles/select2/3/3sub3/3sub3sub2\n"
                    u"+ testfiles/select2/3/3sub2/3sub2sub2\n"
                    u"+ testfiles/select2/3/3sub3\n"
                    u"- testfiles/select2/3/3sub1\n"  # - added to ensure it makes no difference
                    u"testfiles/select2/2/2sub1/2sub1sub3\n"
                    u"testfiles/select2/2/2sub1/2sub1sub2\n"
                    u"+ testfiles/select2/2/2sub1\n"
                    u"testfiles/select2/1/1sub3/1sub3sub2\n"
                    u"testfiles/select2/1/1sub3/1sub3sub1\n"
                    u"testfiles/select2/1/1sub2/1sub2sub3\n"
                    u"+ testfiles/select2/1/1sub2/1sub2sub1\n"
                    u"testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n"
                    u"testfiles/select2/1/1sub1/1sub1sub2\n"
                    u"- testfiles/select2/1/1sub2\n"  # - added to ensure it makes no difference
                    u"+ testfiles/select2/1.py\n"
                    u"+ testfiles/select2/3\n"
                    u"+ testfiles/select2/1\n"
                    u"testfiles/select2/**")

        # Backup the files exactly as in test_exclude_filelist, but with the --progress option
        self.backup(u"full", u"testfiles/select2", options=[u"--exclude-filelist=testfiles/exclude.txt",
                                                          u"--progress"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        # The restored files should match those restored in test_exclude_filelist
        self.assertEqual(restored, self.expected_restored_tree)


class TestIncludeFilelistTest(IncludeExcludeFunctionalTest):
    u"""
    Test --include-filelist using duplicity binary.
    """

    def test_include_filelist(self):
        u"""Test that include filelist works in the basic case"""
        # See test_exclude_filelist above for explanation of what is expected. As this is an include filelist
        # any lines with no +/- modifier should be treated as if they have a +.
        # Create a filelist
        with io.open(u"testfiles/include.txt", u"w") as f:
            f.write(u"testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n"
                    u"- testfiles/select2/3/3sub3/3sub3sub2\n"
                    u"testfiles/select2/3/3sub2/3sub2sub2\n"
                    u"+ testfiles/select2/3/3sub3\n"  # + added to ensure it makes no difference
                    u"- testfiles/select2/3/3sub1\n"
                    u"- testfiles/select2/2/2sub1/2sub1sub3\n"
                    u"- testfiles/select2/2/2sub1/2sub1sub2\n"
                    u"testfiles/select2/2/2sub1\n"
                    u"- testfiles/select2/1/1sub3/1sub3sub2\n"
                    u"- testfiles/select2/1/1sub3/1sub3sub1\n"
                    u"- testfiles/select2/1/1sub2/1sub2sub3\n"
                    u"+ testfiles/select2/1/1sub2/1sub2sub1\n"  # + added to ensure it makes no difference
                    u"- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n"
                    u"- testfiles/select2/1/1sub1/1sub1sub2\n"
                    u"- testfiles/select2/1/1sub2\n"
                    u"testfiles/select2/1.py\n"
                    u"testfiles/select2/3\n"
                    u"testfiles/select2/1\n"
                    u"- testfiles/select2/**")
        self.backup(u"full", u"testfiles/select2", options=[u"--include-filelist=testfiles/include.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_include_filelist_combined_imperfections(self):
        u"""Test that include filelist works with imperfections in the input file"""
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier  (both " and ')
        # Create a filelist
        with io.open(u"testfiles/include.txt", u"w") as f:
            f.write(u"testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n"
                    u"- testfiles/select2/3/3sub3/3sub3sub2\n"
                    u'"testfiles/select2/3/3sub2/3sub2sub2"\n'
                    u"  + testfiles/select2/3/3sub3\n"  # + added to ensure it makes no difference
                    u"- testfiles/select2/3/3sub1\n"
                    u"- testfiles/select2/2/2sub1/2sub1sub3\n"
                    u' - "testfiles/select2/2/2sub1/2sub1sub2"\n'
                    u"testfiles/select2/2/2sub1  \n"
                    u"\n"
                    u"- testfiles/select2/1/1sub3/1sub3sub2\n"
                    u"- testfiles/select2/1/1sub3/1sub3sub1 \n"
                    u"- 'testfiles/select2/1/1sub2/1sub2sub3'\n"
                    u"             \n"
                    u" + testfiles/select2/1/1sub2/1sub2sub1 \n"  # + added to ensure it makes no difference
                    u"- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n"
                    u"  - testfiles/select2/1/1sub1/1sub1sub2  \n"
                    u"# Testing full-line comment\n"
                    u"- testfiles/select2/1/1sub2\n"
                    u"'testfiles/select2/1.py'\n"
                    u"testfiles/select2/3\n"
                    u"        #  Testing another full-line comment      \n"
                    u"testfiles/select2/1\n"
                    u"- testfiles/select2/**")
        self.backup(u"full", u"testfiles/select2", options=[u"--include-filelist=testfiles/include.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_include_filelist_workaround_combined_imperfections_no_wildcards(self):
        u"""Test that include filelist works with imperfections in the input file"""
        # This is a modified version of test_include_filelist that passes, despite Bug #1408411
        # It is still a valid test, it just does not test as many selection features as the above.
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier  (both " and ')
        # Create a filelist
        with io.open(u"testfiles/include.txt", u"w") as f:
            f.write(u"testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n"
                    u"testfiles/select2/3/3sub2/3sub2sub2 \n"
                    u"  + testfiles/select2/3/3sub3\n"  # + added to ensure it makes no difference
                    u" - testfiles/select2/3/3sub1  \n"
                    u"- testfiles/select2/2/2sub1/2sub1sub3\n"
                    u"- testfiles/select2/2/2sub1/2sub1sub2\n"
                    u'"testfiles/select2/2/2sub1"\n'
                    u"   - testfiles/select2/2/2sub3 \n"  # Added because of Bug #1408411
                    u"- testfiles/select2/2/2sub2\n"  # Added because of Bug #1408411
                    u"- 'testfiles/select2/1/1sub3/1sub3sub2'\n"
                    u"\n"
                    u"- testfiles/select2/1/1sub3/1sub3sub1\n"
                    u"- testfiles/select2/1/1sub2/1sub2sub3\n"
                    u'- "testfiles/select2/1/1sub2/1sub2sub2"\n'  # Added because of Bug #1408411
                    u"# This is a full-line comment\n"
                    u"+ testfiles/select2/1/1sub2/1sub2sub1  \n"  # + added to ensure it makes no difference
                    u"- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n"
                    u"          \n"
                    u"- testfiles/select2/1/1sub1/1sub1sub2\n"
                    #  u"- testfiles/select2/1/1sub2\n"  # Commented out because of Bug #1408411
                    u"'testfiles/select2/1.py'\n"
                    u"       # This is another full-line comment, with spaces     \n"
                    u"testfiles/select2/3\n"
                    #  u"- testfiles/select2/2\n" # Commented out because of Bug #1408411
                    u"testfiles/select2/1\n"
                    u'- "testfiles/select2/trailing_space "\n'  # es instead of ea as no wildcard - **
                    u"- testfiles/select2/1.doc")  # es instead of ea as no wildcard - **
        self.backup(u"full", u"testfiles/select2", options=[u"--include-filelist=testfiles/include.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)

    def test_include_globbing_filelist_combined_imperfections(self):
        u"""Test that include globbing filelist works with imperfections in the input file"""
        # Identical to test_include_filelist_combined_imperfections and included to ensure that
        # the deprecated --include-globbing-filelist function works as expected until it is deliberately removed.
        # This is a combined test for speed reasons. The individual imperfections are tested as unittests in
        # unit/test_selection.
        # Imperfections tested are;
        # * Leading space/spaces before the modifier
        # * Trailing space/spaces after the filename (but before the newline)
        # * Blank lines (newline character only)
        # * Line only containing spaces
        # * Full-line comments with # as the first character and with leading/trailing spaces
        # * Unnecessarily quoted filenames with/without modifier  (both " and ')
        # Create a filelist
        with io.open(u"testfiles/include.txt", u"w") as f:
            f.write(u"testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt\n"
                    u"- testfiles/select2/3/3sub3/3sub3sub2\n"
                    u'"testfiles/select2/3/3sub2/3sub2sub2"\n'
                    u"  + testfiles/select2/3/3sub3\n"  # + added to ensure it makes no difference
                    u"- testfiles/select2/3/3sub1\n"
                    u"- testfiles/select2/2/2sub1/2sub1sub3\n"
                    u' - "testfiles/select2/2/2sub1/2sub1sub2"\n'
                    u"testfiles/select2/2/2sub1  \n"
                    u"\n"
                    u"- testfiles/select2/1/1sub3/1sub3sub2\n"
                    u"- testfiles/select2/1/1sub3/1sub3sub1 \n"
                    u"- 'testfiles/select2/1/1sub2/1sub2sub3'\n"
                    u"             \n"
                    u" + testfiles/select2/1/1sub2/1sub2sub1 \n"  # + added to ensure it makes no difference
                    u"- testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt\n"
                    u"  - testfiles/select2/1/1sub1/1sub1sub2  \n"
                    u"# Testing full-line comment\n"
                    u"- testfiles/select2/1/1sub2\n"
                    u"'testfiles/select2/1.py'\n"
                    u"testfiles/select2/3\n"
                    u"        #  Testing another full-line comment      \n"
                    u"testfiles/select2/1\n"
                    u"- testfiles/select2/**")
        self.backup(u"full", u"testfiles/select2", options=[u"--include-globbing-filelist=testfiles/include.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)


class TestIncludeExcludedForContents(IncludeExcludeFunctionalTest):
    u""" Test to check that folders that are excluded are included if they contain includes of higher priority.
     Exhibits the issue reported in Bug #1408411 (https://bugs.launchpad.net/duplicity/+bug/1408411). """

    def write_filelist(self, filelist_name):
        u"""Used by the below tests to write the filelist"""
        assert filelist_name is not None
        with io.open(filelist_name, u"w") as f:
            f.write(u"+ testfiles/select/1/2/1\n"
                    u"- testfiles/select/1/2\n"
                    u"- testfiles/select/1/1\n"
                    u"- testfiles/select/1/3")

    def restore_and_check(self):
        u"""Restores the backup and compares to what was expected (based on the filelist in write_filelist)"""
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"2"], [u"1"]])

    def test_commandline_include_exclude(self):
        u"""test an excluded folder is included for included contents when using commandline includes and excludes"""
        self.backup(u"full", u"testfiles/select/1",
                    options=[u"--include", u"testfiles/select/1/2/1",
                             u"--exclude", u"testfiles/select/1/2",
                             u"--exclude", u"testfiles/select/1/1",
                             u"--exclude", u"testfiles/select/1/3"])
        self.restore_and_check()

    def test_include_globbing_filelist(self):
        u"""test an excluded folder is included for included contents with an include-globbing-filelist """
        # Deprecated, but include for now to ensure it keeps working until it is deliberately removed.
        self.write_filelist(u"testfiles/include.txt")
        self.backup(u"full", u"testfiles/select/1", options=[u"--include-globbing-filelist=testfiles/include.txt"])
        self.restore_and_check()

    def test_exclude_globbing_filelist(self):
        u"""test an excluded folder is included for included contents with an exclude-globbing-filelist """
        # Deprecated, but include for now to ensure it keeps working until it is deliberately removed.
        self.write_filelist(u"testfiles/exclude.txt")
        self.backup(u"full", u"testfiles/select/1", options=[u"--exclude-globbing-filelist=testfiles/exclude.txt"])
        self.restore_and_check()

    def test_include_filelist(self):
        u"""test an excluded folder is included for included contents with an include-filelist (non-globbing) """
        # Regression test for Bug #1408411 (https://bugs.launchpad.net/duplicity/+bug/1408411)
        self.write_filelist(u"testfiles/include.txt")
        self.backup(u"full", u"testfiles/select/1", options=[u"--include-filelist=testfiles/include.txt"])
        self.restore_and_check()

    def test_exclude_filelist(self):
        u"""test an excluded folder is included for included contents with an exclude-filelist  (non-globbing) """
        # Regression test for Bug #1408411 (https://bugs.launchpad.net/duplicity/+bug/1408411)
        self.write_filelist(u"testfiles/exclude.txt")
        self.backup(u"full", u"testfiles/select/1", options=[u"--exclude-filelist=testfiles/exclude.txt"])
        self.restore_and_check()


class TestAsterisks(IncludeExcludeFunctionalTest):
    u""" Test to check that asterisks work as expected
     Exhibits the issue reported in Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371).
     See the unit tests for more granularity on the issue."""

    def restore_and_check(self):
        u"""Restores the backup and compares to what is expected."""
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"2"], [u"1"]])

    def test_exclude_filelist_asterisks_none(self):
        u"""Basic exclude filelist."""
        with io.open(u"testfiles/filelist.txt", u"w") as f:
            f.write(u"+ testfiles/select/1/2/1\n"
                    u"- testfiles/select/1/2\n"
                    u"- testfiles/select/1/1\n"
                    u"- testfiles/select/1/3")
        self.backup(u"full", u"testfiles/select/1", options=[u"--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_asterisks_single(self):
        u"""Exclude filelist with asterisks replacing folders."""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        with io.open(u"testfiles/filelist.txt", u"w") as f:
            f.write(u"+ */select/1/2/1\n"
                    u"- */select/1/2\n"
                    u"- testfiles/*/1/1\n"
                    u"- */*/1/3")
        self.backup(u"full", u"testfiles/select/1", options=[u"--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_asterisks_double_asterisks(self):
        u"""Exclude filelist with double asterisks replacing folders."""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        with io.open(u"testfiles/filelist.txt", u"w") as f:
            f.write(u"+ **/1/2/1\n"
                    u"- **/1/2\n"
                    u"- **/select/1/1\n"
                    u"- testfiles/select/1/3")
        self.backup(u"full", u"testfiles/select/1", options=[u"--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_commandline_asterisks_single_excludes_only(self):
        u"""test_commandline_include_exclude with single asterisks on exclude lines."""
        self.backup(u"full", u"testfiles/select/1",
                    options=[u"--include", u"testfiles/select/1/2/1",
                             u"--exclude", u"testfiles/*/1/2",
                             u"--exclude", u"*/select/1/1",
                             u"--exclude", u"*/select/1/3"])
        self.restore_and_check()

    def test_commandline_asterisks_single_both(self):
        u"""test_commandline_include_exclude with single asterisks on both exclude and include lines."""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.backup(u"full", u"testfiles/select/1",
                    options=[u"--include", u"*/select/1/2/1",
                             u"--exclude", u"testfiles/*/1/2",
                             u"--exclude", u"*/select/1/1",
                             u"--exclude", u"*/select/1/3"])
        self.restore_and_check()

    def test_commandline_asterisks_double_exclude_only(self):
        u"""test_commandline_include_exclude with double asterisks on exclude lines."""
        self.backup(u"full", u"testfiles/select/1",
                    options=[u"--include", u"testfiles/select/1/2/1",
                             u"--exclude", u"**/1/2",
                             u"--exclude", u"**/1/1",
                             u"--exclude", u"**/1/3"])
        self.restore_and_check()

    def test_commandline_asterisks_double_both(self):
        u"""test_commandline_include_exclude with double asterisks on both exclude and include lines."""
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        self.backup(u"full", u"testfiles/select/1",
                    options=[u"--include", u"**/1/2/1",
                             u"--exclude", u"**/1/2",
                             u"--exclude", u"**/1/1",
                             u"--exclude", u"**/1/3"])
        self.restore_and_check()

    def test_single_and_double_asterisks(self):
        u"""This compares a backup using --include-globbing-filelist with a single and double *."""
        with io.open(u"testfiles/filelist.txt", u"w") as f:
            f.write(u"+ testfiles/select2/*\n"
                    u"- testfiles/select")
        self.backup(u"full", u"testfiles/", options=[u"--include-globbing-filelist=testfiles/filelist.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir + u"/select2")
        with io.open(u"testfiles/filelist2.txt", u"w") as f:
            f.write(u"+ testfiles/select2/**\n"
                    u"- testfiles/select")
        self.backup(u"full", u"testfiles/", options=[u"--include-globbing-filelist=testfiles/filelist2.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored2 = self.directory_tree_to_list_of_lists(restore_dir + u"/select2")
        self.assertEqual(restored, restored2)

    def test_single_and_double_asterisks_includes_excludes(self):
        u"""This compares a backup using --includes/--excludes with a single and double *."""
        self.backup(u"full", u"testfiles/",
                    options=[u"--include", u"testfiles/select2/*",
                             u"--exclude", u"testfiles/select"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir + u"/select2")
        self.backup(u"full", u"testfiles/",
                    options=[u"--include", u"testfiles/select2/**",
                             u"--exclude", u"testfiles/select"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored2 = self.directory_tree_to_list_of_lists(restore_dir + u"/select2")
        self.assertEqual(restored, restored2)


class TestTrailingSlash(IncludeExcludeFunctionalTest):
    u""" Test to check that a trailing slash works as expected
     Exhibits the issue reported in Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)."""

    def restore_and_check(self):
        u"""Restores the backup and compares to what is expected."""
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"2"], [u"1"]])

    def test_exclude_filelist_trailing_slashes(self):
        u"""test_exclude_filelist_asterisks_none with trailing slashes."""
        with io.open(u"testfiles/filelist.txt", u"w") as f:
            f.write(u"+ testfiles/select/1/2/1/\n"
                    u"- testfiles/select/1/2/\n"
                    u"- testfiles/select/1/1/\n"
                    u"- testfiles/select/1/3/")
        self.backup(u"full", u"testfiles/select/1", options=[u"--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_trailing_slashes_single_wildcards_excludes(self):
        u"""test_exclude_filelist_trailing_slashes with single wildcards in excludes."""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)
        with io.open(u"testfiles/filelist.txt", u"w") as f:
            f.write(u"+ testfiles/select/1/2/1/\n"
                    u"- */select/1/2/\n"
                    u"- testfiles/*/1/1/\n"
                    u"- */*/1/3/")
        self.backup(u"full", u"testfiles/select/1", options=[u"--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_trailing_slashes_double_wildcards_excludes(self):
        u"""test_exclude_filelist_trailing_slashes with double wildcards in excludes."""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)
        with io.open(u"testfiles/filelist.txt", u"w") as f:
            f.write(u"+ testfiles/select/1/2/1/\n"
                    u"- **/1/2/\n"
                    u"- **/1/1/\n"
                    u"- **/1/3/")
        self.backup(u"full", u"testfiles/select/1", options=[u"--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_trailing_slashes_double_wildcards_excludes_2(self):
        u"""second test_exclude_filelist_trailing_slashes with double wildcards in excludes."""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482) and
        # Regression test for Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371)
        with io.open(u"testfiles/filelist.txt", u"w") as f:
            f.write(u"+ **/1/2/1/\n"
                    u"- **/1/2/\n"
                    u"- **/1/1/\n"
                    u"- **/1/3/")
        self.backup(u"full", u"testfiles/select/1", options=[u"--exclude-filelist=testfiles/filelist.txt"])
        self.restore_and_check()

    def test_exclude_filelist_trailing_slashes_wildcards(self):
        u"""test_commandline_asterisks_single_excludes_only with trailing slashes."""
        # Regression test for Bug #932482 (https://bugs.launchpad.net/duplicity/+bug/932482)
        self.backup(u"full", u"testfiles/select/1",
                    options=[u"--include", u"testfiles/select/1/2/1/",
                             u"--exclude", u"testfiles/*/1/2/",
                             u"--exclude", u"*/select/1/1/",
                             u"--exclude", u"*/select/1/3/"])
        self.restore_and_check()


class TestTrailingSlash2(IncludeExcludeFunctionalTest):
    u""" This tests the behaviour of globbing strings with a trailing slash"""
    # See Bug #1479545 (https://bugs.launchpad.net/duplicity/+bug/1479545)

    def test_no_trailing_slash(self):
        u""" Test that including 1.py works as expected"""
        self.backup(u"full", u"testfiles/select2",
                    options=[u"--include", u"testfiles/select2/1.py",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"1.py"]])

    def test_trailing_slash(self):
        u""" Test that globs with a trailing slash only match directories"""
        # Regression test for Bug #1479545
        # (https://bugs.launchpad.net/duplicity/+bug/1479545)
        self.backup(u"full", u"testfiles/select2",
                    options=[u"--include", u"testfiles/select2/1.py/",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])

    def test_include_files_not_subdirectories(self):
        u""" Test that a trailing slash glob followed by a * glob only matches
        files and not subdirectories"""
        self.backup(u"full", u"testfiles/select2",
                    options=[u"--exclude", u"testfiles/select2/*/",
                             u"--include", u"testfiles/select2/*",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"1.doc", u"1.py"]])

    def test_include_subdirectories_not_files(self):
        u""" Test that a trailing slash glob only matches directories"""
        self.backup(u"full", u"testfiles/select2",
                    options=[u"--include", u"testfiles/select2/1/1sub1/**/",
                             u"--exclude", u"testfiles/select2/1/1sub1/**",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"1"], [u"1sub1"],
                                    [u"1sub1sub1", u"1sub1sub2", u"1sub1sub3"]])


class TestGlobbingReplacement(IncludeExcludeFunctionalTest):
    u""" This tests the behaviour of the extended shell globbing pattern replacement functions."""
    # See the manual for a description of behaviours, but in summary:
    # * can be expanded to any string of characters not containing "/"
    # ? expands to any character except "/" and
    # [...] expands to a single character of those characters specified (ranges are acceptable).
    # The new special pattern, **, expands to any string of characters whether or not it contains "/".
    # Furthermore, if the pattern starts with "ignorecase:" (case insensitive), then this prefix will be
    # removed and any character in the string can be replaced with an upper- or lowercase version of itself.

    def test_globbing_replacement_in_includes(self):
        u""" Test behaviour of the extended shell globbing pattern replacement functions in both include and exclude"""
        # Identical to test_include_exclude_basic with globbing characters added to both include and exclude lines
        # Exhibits the issue reported in Bug #884371 (https://bugs.launchpad.net/duplicity/+bug/884371).
        # See above and the unit tests for more granularity on the issue.
        self.backup(u"full", u"testfiles/select2",
                    options=[u"--include", u"testfiles/select2/**/3sub3sub2/3sub3su?2_file.txt",  # Note ** and ? added
                             u"--exclude", u"testfiles/select2/*/3s*1",  # Note * added in both directory and filename
                             u"--exclude", u"testfiles/select2/**/2sub1sub3",  # Note ** added
                             u"--exclude", u"ignorecase:testfiles/select2/2/2sub1/2Sub1Sub2",  # Note ignorecase added
                             u"--include", u"ignorecase:testfiles/sel[w,u,e,q]ct2/2/2S?b1",    # Note ignorecase, [] and
                             # ? added
                             u"--exclude", u"testfiles/select2/1/1sub3/1s[w,u,p,q]b3sub2",  # Note [] added
                             u"--exclude", u"testfiles/select2/1/1sub[1-4]/1sub3sub1",  # Note [range] added
                             u"--include", u"testfiles/select2/*/1sub2/1s[w,u,p,q]b2sub1",  # Note * and [] added
                             u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub3/1su?1sub3_file.txt",  # Note ? added
                             u"--exclude", u"testfiles/select2/1/1*1/1sub1sub2",  # Note * added
                             u"--exclude", u"testfiles/select2/1/1sub2",
                             u"--include", u"testfiles/select[2-4]/*.py",  # Note * and [range] added
                             u"--include", u"testfiles/*2/3",  # Note * added
                             u"--include", u"**/select2/1",  # Note ** added
                             u"--exclude", u"testfiles/select2/**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)


class TestExcludeIfPresent(IncludeExcludeFunctionalTest):
    u""" This tests the behaviour of duplicity's --exclude-if-present option"""

    def test_exclude_if_present_baseline(self):
        u""" Test that duplicity normally backs up files"""
        with io.open(u"testfiles/select2/1/1sub1/1sub1sub1/.nobackup", u"w") as tag:
            tag.write(u"Files in this folder should not be backed up.")
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--include", u"testfiles/select2/1/1sub1/1sub1sub1/*",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"1sub1sub1"],
                                    [u".nobackup", u"1sub1sub1_file.txt"]])

    def test_exclude_if_present_excludes(self):
        u""" Test that duplicity excludes files with relevant tag"""
        with io.open(u"testfiles/select2/1/1sub1/1sub1sub1/.nobackup", u"w") as tag:
            tag.write(u"Files in this folder should not be backed up.")
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--exclude-if-present", u".nobackup",
                             u"--include", u"testfiles/select2/1/1sub1/1sub1sub1/*",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])

    def test_exclude_if_present_excludes_2(self):
        u""" Test that duplicity excludes files with relevant tag"""
        with io.open(u"testfiles/select2/1/1sub1/1sub1sub1/EXCLUDE.tag", u"w") as tag:
            tag.write(u"Files in this folder should also not be backed up.")
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--exclude-if-present", u"EXCLUDE.tag",
                             u"--include", u"testfiles/select2/1/1sub1/1sub1sub1/*",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])


class TestLockedFoldersNoError(IncludeExcludeFunctionalTest):
    u""" This tests that inaccessible folders do not cause an error"""

    @unittest.skipUnless(platform.platform().startswith(u"Linux"),
                         u"Skip on non-Linux systems")
    def test_locked_baseline(self):
        u""" Test no error if locked in path but excluded"""
        folder_to_lock = u"testfiles/select2/1/1sub1/1sub1sub3"
        initial_mode = os.stat(folder_to_lock).st_mode
        os.chmod(folder_to_lock, 0o0000)
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--include", u"testfiles/select2/1/1sub1/1sub1sub1/*",
                             u"--exclude", u"**"])
        os.chmod(folder_to_lock, initial_mode)
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"1sub1sub1"],
                                    [u"1sub1sub1_file.txt"]])

    @unittest.skipUnless(platform.platform().startswith(u"Linux"),
                         u"Skip on non-Linux systems")
    def test_locked_excl_if_present(self):
        u""" Test no error if excluded locked with --exclude-if-present"""
        # Regression test for Bug #1620085
        # https://bugs.launchpad.net/duplicity/+bug/1620085
        folder_to_lock = u"testfiles/select2/1/1sub1/1sub1sub3"
        initial_mode = os.stat(folder_to_lock).st_mode
        os.chmod(folder_to_lock, 0o0000)
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--exclude-if-present", u"EXCLUDE.tag",
                             u"--include", u"testfiles/select2/1/1sub1/1sub1sub1/*",
                             u"--exclude", u"**"])
        os.chmod(folder_to_lock, initial_mode)
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"1sub1sub1"],
                                    [u"1sub1sub1_file.txt"]])


class TestFolderIncludesFiles(IncludeExcludeFunctionalTest):
    u""" This tests that including a folder includes the files within it"""
    # https://bugs.launchpad.net/duplicity/+bug/1624725

    def test_includes_files(self):
        u"""This tests that including a folder includes the files within it"""
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--include", u"testfiles/select2/1/1sub1/1sub1sub1",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"1sub1sub1"],
                                    [u"1sub1sub1_file.txt"]])

    def test_includes_files_trailing_slash(self):
        u"""This tests that including a folder includes the files within it"""
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--include", u"testfiles/select2/1/1sub1/1sub1sub1/",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"1sub1sub1"],
                                    [u"1sub1sub1_file.txt"]])

    def test_includes_files_trailing_slash_globbing_chars(self):
        u"""Tests folder includes with globbing char and /"""
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--include", u"testfiles/s?lect2/1/1sub1/1sub1sub1/",
                             u"--exclude", u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"1sub1sub1"],
                                    [u"1sub1sub1_file.txt"]])

    def test_excludes_files_no_trailing_slash(self):
        u"""This tests that excluding a folder excludes the files within it"""
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub1",
                             u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub2",
                             u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub3",
                             u"--include", u"testfiles/select2/1/1sub1/1sub1**",
                             u"--exclude", u"testfiles/select2/1/1sub1/irrelevant.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])

    def test_excludes_files_trailing_slash(self):
        u"""Excluding a folder excludes the files within it, if ends with /"""
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub1/",
                             u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub2/",
                             u"--exclude", u"testfiles/select2/1/1sub1/1sub1sub3/",
                             u"--include", u"testfiles/select2/1/1sub1/1sub1**",
                             u"--exclude", u"testfiles/select2/1/1sub1/irrelevant.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])

    def test_excludes_files_trailing_slash_globbing_chars(self):
        u"""Tests folder excludes with globbing char and /"""
        self.backup(u"full", u"testfiles/select2/1/1sub1",
                    options=[u"--exclude", u"testfiles/sel?ct2/1/1sub1/1sub1sub1/",
                             u"--exclude", u"testfiles/sel[e,f]ct2/1/1sub1/1sub1sub2/",
                             u"--exclude", u"testfiles/sel*t2/1/1sub1/1sub1sub3/",
                             u"--include", u"testfiles/select2/1/1sub1/1sub1**",
                             u"--exclude", u"testfiles/select2/1/1sub1/irrelevant.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [])


class TestAbsolutePaths(IncludeExcludeFunctionalTest):
    u""" Tests include/exclude options with absolute paths"""

    def test_absolute_paths_non_globbing(self):
        u""" Test --include and --exclude work with absolute paths"""
        self.backup(u"full", os.path.abspath(u"testfiles/select2"),
                    options=[u"--include", os.path.abspath(u"testfiles/select2/3/3sub3/3sub3sub2/3sub3sub2_file.txt"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/3/3sub3/3sub3sub2"),
                             u"--include", os.path.abspath(u"testfiles/select2/3/3sub2/3sub2sub2"),
                             u"--include", os.path.abspath(u"testfiles/select2/3/3sub3"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/3/3sub1"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/2/2sub1/2sub1sub3"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/2/2sub1/2sub1sub2"),
                             u"--include", os.path.abspath(u"testfiles/select2/2/2sub1"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/1/1sub3/1sub3sub2"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/1/1sub3/1sub3sub1"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/1/1sub2/1sub2sub3"),
                             u"--include", os.path.abspath(u"testfiles/select2/1/1sub2/1sub2sub1"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/1/1sub1/1sub1sub3/1sub1sub3_file.txt"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/1/1sub1/1sub1sub2"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/1/1sub2"),
                             u"--include", os.path.abspath(u"testfiles/select2/1.py"),
                             u"--include", os.path.abspath(u"testfiles/select2/3"),
                             u"--include", os.path.abspath(u"testfiles/select2/1"),
                             u"--exclude", os.path.abspath(u"testfiles/select2/**")])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, self.expected_restored_tree)


@unittest.skipUnless(platform.platform().startswith(u"Linux"), u"Skip on non-Linux systems")
@unittest.skipUnless(sys.getfilesystemencoding().upper() == u"UTF-8", u"Skip on non-UTF-8 systems")
@unittest.skipIf(sys.version_info[:2] < (3, 7), u"Skip on bad unicode handling")
class TestUnicode(IncludeExcludeFunctionalTest):
    u""" Tests include/exclude options with unicode paths"""

    def test_unicode_paths_non_globbing(self):
        u""" Test --include and --exclude work with unicode paths"""
        self.backup(u"full", u"testfiles/select-unicode",
                    options=[u"--exclude", u"testfiles/select-unicode/прыклад/пример/例/Παράδειγμα/उदाहरण.txt",
                             u"--exclude", u"testfiles/select-unicode/прыклад/пример/例/Παράδειγμα/דוגמא.txt",
                             u"--exclude", u"testfiles/select-unicode/прыклад/пример/例/მაგალითი/",
                             u"--include", u"testfiles/select-unicode/прыклад/пример/例/",
                             u"--exclude", u"testfiles/select-unicode/прыклад/пример/",
                             u"--include", u"testfiles/select-unicode/прыклад/",
                             u"--include", u"testfiles/select-unicode/օրինակ.txt",
                             u"--exclude", u"testfiles/select-unicode/**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"прыклад", u"օրինակ.txt"],
                                    [u"пример", u"উদাহরণ"], [u"例"], [u"Παράδειγμα"], [u"ઉદાહરણ.log"]])

    def test_unicode_paths_asterisks(self):
        u""" Test --include and --exclude work with unicode paths and globs containing * and **"""
        p = u"testfiles/select-unicode/"
        self.backup(u"full", u"testfiles/select-unicode",
                    options=[u"--exclude", p + u"прыклад/пример/例/Παρά*ειγμα/उदाहरण.txt",  # Note *
                             u"--exclude", p + u"прыклад/пример/例/Παράδειγμα/דוגמא.txt",
                             u"--exclude", p + u"прыклад/пример/例/მაგალითი/",
                             u"--include", p + u"пр**/例/",  # Note **
                             u"--exclude", p + u"прыклад/пример/",
                             u"--include", p + u"прыкла*/",  # Note *
                             u"--include", p + u"օր*ակ.txt",  # Note *
                             u"--exclude", p + u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"прыклад", u"օրինակ.txt"],
                                    [u"пример", u"উদাহরণ"], [u"例"], [u"Παράδειγμα"], [u"ઉદાહરણ.log"]])

    def test_unicode_paths_square_brackets(self):
        u""" Test --include and --exclude work with unicode paths with character options in []s and [!]s"""
        p = u"testfiles/select-unicode/"
        self.backup(u"full", u"testfiles/select-unicode",
                    options=[u"--exclude", p + u"прыклад/пример/例/Παράδειγμα/उदाहरण.txt",
                             u"--exclude", p + u"пры[к,и,р]лад/пример/例/Παράδειγμα/דוגמא.txt",
                             u"--exclude", p + u"прыклад/пр[!a,b,c]мер/例/მაგალითი/",
                             u"--include", p + u"прыклад/при[g,м,д]ер/例/",
                             u"--exclude", p + u"прыклад/пример/",
                             u"--include", p + u"прыклад/",
                             u"--include", p + u"օրինակ.txt",
                             u"--exclude", p + u"**"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"прыклад", u"օրինակ.txt"],
                                    [u"пример", u"উদাহরণ"], [u"例"], [u"Παράδειγμα"], [u"ઉદાહરણ.log"]])

    def test_unicode_filelist(self):
        u"""Test that exclude filelist works with unicode filenames"""
        # As this is an exclude filelist any lines with no +/- modifier should be treated as if they have a -.
        path = u"testfiles/select-unicode/"
        # Create a filelist
        with io.open(u"testfiles/exclude.txt", u"w", encoding=u"UTF-8") as f:
            f.write(u"- " + path + u"прыклад/пример/例/Παράδειγμα/उदाहरण.txt\n"
                    u"- " + path + u"прыклад/пример/例/Παράδειγμα/דוגמא.txt\n"
                    u"- " + path + u"прыклад/пример/例/მაგალითი/\n"
                    u"+ " + path + u"прыклад/пример/例/\n"
                    u"- " + path + u"прыклад/пример/\n"
                    u"+ " + path + u"прыклад/\n"
                    u"+ " + path + u"օրինակ.txt\n"
                    u"- " + path + u"**")
        self.backup(u"full", path, options=[u"--exclude-filelist=testfiles/exclude.txt"])
        self.restore()
        restore_dir = u"testfiles/restore_out"
        restored = self.directory_tree_to_list_of_lists(restore_dir)
        self.assertEqual(restored, [[u"прыклад", u"օրինակ.txt"],
                                    [u"пример", u"উদাহরণ"], [u"例"], [u"Παράδειγμα"], [u"ઉદાહરણ.log"]])

if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2015 Yigal Asnis
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# It is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from builtins import next
from builtins import str

import os

from duplicity import log
from duplicity import util
from duplicity.errors import BackendException
import duplicity.backend


class PyDriveBackend(duplicity.backend.Backend):
    u"""Connect to remote store using PyDrive API"""

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)
        try:
            import httplib2
            from apiclient.discovery import build
        except ImportError as e:
            raise BackendException(u"""\
PyDrive backend requires PyDrive and Google API client installation.
Please read the manpage for setup details.
Exception: %s""" % str(e))

        # Shared Drive ID specified as a query parameter in the backend URL.
        # Example: pydrive://developer.gserviceaccount.com/target-folder/?driveID=<SHARED DRIVE ID>
        self.api_params = {}
        self.shared_drive_id = None
        if u'driveID' in parsed_url.query_args:
            self.shared_drive_id = parsed_url.query_args[u'driveID'][0]
            self.api_params = {u'corpora': u'teamDrive',
                               u'teamDriveId': self.shared_drive_id,
                               u'includeTeamDriveItems': True,
                               u'supportsTeamDrives': True}

        try:
            from pydrive2.auth import GoogleAuth
            from pydrive2.drive import GoogleDrive
            from pydrive2.files import ApiRequestError, FileNotUploadedError
        except ImportError as e:
            try:
                from pydrive.auth import GoogleAuth
                from pydrive.drive import GoogleDrive
                from pydrive.files import ApiRequestError, FileNotUploadedError
            except ImportError as e:
                raise BackendException(u"""\
PyDrive backend requires PyDrive installation.  Please read the manpage for setup details.
Exception: %s""" % str(e))

        # let user get by with old client while he can
        try:
            from oauth2client.client import SignedJwtAssertionCredentials
            self.oldClient = True
        except:
            from oauth2client.service_account import ServiceAccountCredentials
            from oauth2client import crypt
            self.oldClient = False

        if u'GOOGLE_DRIVE_ACCOUNT_KEY' in os.environ:
            account_key = os.environ[u'GOOGLE_DRIVE_ACCOUNT_KEY']
            if self.oldClient:
                credentials = SignedJwtAssertionCredentials(parsed_url.username +
                                                            u'@' + parsed_url.hostname,
                                                            account_key,
                                                            scopes=u'https://www.googleapis.com/auth/drive')
            else:
                signer = crypt.Signer.from_string(account_key)
                credentials = ServiceAccountCredentials(parsed_url.username + u'@' + parsed_url.hostname, signer,
                                                        scopes=u'https://www.googleapis.com/auth/drive')
            credentials.authorize(httplib2.Http())
            gauth = GoogleAuth(http_timeout=60)
            gauth.credentials = credentials
        elif u'GOOGLE_DRIVE_SETTINGS' in os.environ:
            gauth = GoogleAuth(settings_file=os.environ[u'GOOGLE_DRIVE_SETTINGS'], http_timeout=60)
            gauth.CommandLineAuth()
        elif (u'GOOGLE_SECRETS_FILE' in os.environ and u'GOOGLE_CREDENTIALS_FILE' in os.environ):
            gauth = GoogleAuth(http_timeout=60)
            gauth.LoadClientConfigFile(os.environ[u'GOOGLE_SECRETS_FILE'])
            gauth.LoadCredentialsFile(os.environ[u'GOOGLE_CREDENTIALS_FILE'])
            if gauth.credentials is None:
                gauth.CommandLineAuth()
            elif gauth.access_token_expired:
                gauth.Refresh()
            else:
                gauth.Authorize()
            gauth.SaveCredentialsFile(os.environ[u'GOOGLE_CREDENTIALS_FILE'])
        else:
            raise BackendException(
                u'GOOGLE_DRIVE_ACCOUNT_KEY or GOOGLE_DRIVE_SETTINGS environment '
                u'variable not set. Please read the manpage to fix.')
        self.drive = GoogleDrive(gauth)

        if self.shared_drive_id:
            parent_folder_id = self.shared_drive_id
        else:
            # Dirty way to find root folder id
            file_list = self.drive.ListFile({u'q': u"'Root' in parents and trashed=false"}).GetList()
            if file_list:
                parent_folder_id = file_list[0][u'parents'][0][u'id']
            else:
                file_in_root = self.drive.CreateFile({u'title': u'i_am_in_root'})
                file_in_root.Upload()
                parent_folder_id = file_in_root[u'parents'][0][u'id']
                file_in_root.Delete()

        # Fetch destination folder entry and create hierarchy if required.
        folder_names = parsed_url.path.split(u'/')
        for folder_name in folder_names:
            if not folder_name:
                continue
            list_file_args = {u'q': u"'" + parent_folder_id +
                              u"' in parents and trashed=false"}
            list_file_args.update(self.api_params)
            file_list = self.drive.ListFile(list_file_args).GetList()
            folder = next((item for item in file_list if item[u'title'] == folder_name and
                           item[u'mimeType'] == u'application/vnd.google-apps.folder'), None)
            if folder is None:
                create_file_args = {u'title': folder_name,
                                    u'mimeType': u"application/vnd.google-apps.folder",
                                    u'parents': [{u'id': parent_folder_id}]}
                create_file_args[u'parents'][0].update(self.api_params)
                create_file_args.update(self.api_params)
                folder = self.drive.CreateFile(create_file_args)
                if self.shared_drive_id:
                    folder.Upload(param={u'supportsTeamDrives': True})
                else:
                    folder.Upload()
            parent_folder_id = folder[u'id']
        self.folder = parent_folder_id
        self.id_cache = {}

    def file_by_name(self, filename):
        try:
            from pydrive2.files import ApiRequestError  # pylint: disable=import-error
        except ImportError:
            from pydrive.files import ApiRequestError  # pylint: disable=import-error

        filename = util.fsdecode(filename)  # PyDrive deals with unicode filenames

        if filename in self.id_cache:
            # It might since have been locally moved, renamed or deleted, so we
            # need to validate the entry.
            file_id = self.id_cache[filename]
            drive_file = self.drive.CreateFile({u'id': file_id})
            try:
                if drive_file[u'title'] == filename and not drive_file[u'labels'][u'trashed']:
                    for parent in drive_file[u'parents']:
                        if parent[u'id'] == self.folder:
                            log.Info(u"PyDrive backend: found file '%s' with id %s in ID cache" %
                                     (filename, file_id))
                            return drive_file
            except ApiRequestError as error:
                # A 404 occurs if the ID is no longer valid
                if error.args[0].resp.status != 404:
                    raise
            # If we get here, the cache entry is invalid
            log.Info(u"PyDrive backend: invalidating '%s' (previously ID %s) from ID cache" %
                     (filename, file_id))
            del self.id_cache[filename]

        # Not found in the cache, so use directory listing. This is less
        # reliable because there is no strong consistency.
        q = u"title='%s' and '%s' in parents and trashed=false" % (filename, self.folder)
        fields = u'items(title,id,fileSize,downloadUrl,exportLinks),nextPageToken'
        list_file_args = {u'q': q, u'fields': fields}
        list_file_args.update(self.api_params)
        flist = self.drive.ListFile(list_file_args).GetList()
        if len(flist) > 1:
            log.FatalError(_(u"PyDrive backend: multiple files called '%s'.") % (filename,))
        elif flist:
            file_id = flist[0][u'id']
            self.id_cache[filename] = flist[0][u'id']
            log.Info(u"PyDrive backend: found file '%s' with id %s on server, "
                     u"adding to cache" % (filename, file_id))
            return flist[0]
        log.Info(u"PyDrive backend: file '%s' not found in cache or on server" %
                 (filename,))
        return None

    def id_by_name(self, filename):
        drive_file = self.file_by_name(filename)
        if drive_file is None:
            return u''
        else:
            return drive_file[u'id']

    def _put(self, source_path, remote_filename):
        remote_filename = util.fsdecode(remote_filename)
        drive_file = self.file_by_name(remote_filename)
        if drive_file is None:
            # No existing file, make a new one
            create_file_args = {u'title': remote_filename,
                                u'parents': [{u"kind": u"drive#fileLink",
                                             u"id": self.folder}]}
            create_file_args[u'parents'][0].update(self.api_params)
            drive_file = self.drive.CreateFile(create_file_args)
            log.Info(u"PyDrive backend: creating new file '%s'" % (remote_filename,))
        else:
            log.Info(u"PyDrive backend: replacing existing file '%s' with id '%s'" % (
                remote_filename, drive_file[u'id']))
        drive_file.SetContentFile(util.fsdecode(source_path.name))
        if self.shared_drive_id:
            drive_file.Upload(param={u'supportsTeamDrives': True})
        else:
            drive_file.Upload()
        self.id_cache[remote_filename] = drive_file[u'id']

    def _get(self, remote_filename, local_path):
        drive_file = self.file_by_name(remote_filename)
        drive_file.GetContentFile(util.fsdecode(local_path.name))

    def _list(self):
        list_file_args = {
            u'q': u"'" + self.folder + u"' in parents and trashed=false",
            u'fields': u'items(title,id),nextPageToken'}
        list_file_args.update(self.api_params)
        drive_files = self.drive.ListFile(list_file_args).GetList()
        filenames = set(item[u'title'] for item in drive_files)
        # Check the cache as well. A file might have just been uploaded but
        # not yet appear in the listing.
        # Note: do not use iterkeys() here, because file_by_name will modify
        # the cache if it finds invalid entries.
        for filename in list(self.id_cache.keys()):
            if (filename not in filenames) and (self.file_by_name(filename) is not None):
                filenames.add(filename)
        return list(filenames)

    def _delete(self, filename):
        file_id = self.id_by_name(filename)
        if file_id == u'':
            log.Warn(u"File '%s' does not exist while trying to delete it" % (util.fsdecode(filename),))
        elif self.shared_drive_id:
            self.drive.auth.service.files().delete(fileId=file_id, param={u'supportsTeamDrives': True}).execute()
        else:
            self.drive.auth.service.files().delete(fileId=file_id).execute()

    def _query(self, filename):
        drive_file = self.file_by_name(filename)
        if drive_file is None:
            size = -1
        else:
            size = int(drive_file[u'fileSize'])
        return {u'size': size}

    def _error_code(self, operation, error):  # pylint: disable=unused-argument
        try:
            from pydrive2.files import ApiRequestError, FileNotUploadedError  # pylint: disable=import-error
        except ImportError:
            from pydrive.files import ApiRequestError, FileNotUploadedError  # pylint: disable=import-error
        if isinstance(error, FileNotUploadedError):
            return log.ErrorCode.backend_not_found
        elif isinstance(error, ApiRequestError):
            return log.ErrorCode.backend_permission_denied
        return log.ErrorCode.backend_error


duplicity.backend.register_backend(u'pydrive', PyDriveBackend)
u""" pydrive is an alternate way to access gdocs """
duplicity.backend.register_backend(u'pydrive+gdocs', PyDriveBackend)
u""" register pydrive as the default way to access gdocs """
duplicity.backend.register_backend(u'gdocs', PyDriveBackend)

duplicity.backend.uses_netloc.extend([u'pydrive', u'pydrive+gdocs', u'gdocs'])



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
# Copyright 2010 Marcel Pennewiss <opensource@pennewiss.de>
# Copyright 2014 Edgar Soldin
#                 - webdav, fish, sftp support
#                 - https cert verification switches
#                 - debug output
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from future import standard_library
standard_library.install_aliases()

import os
import os.path
import re
import urllib.request  # pylint: disable=import-error
import urllib.parse  # pylint: disable=import-error
import urllib.error  # pylint: disable=import-error
try:
    from shlex import quote as cmd_quote
except ImportError:
    from pipes import quote as cmd_quote

import duplicity.backend
from duplicity import config
from duplicity import log
from duplicity import tempdir
from duplicity import util


class LFTPBackend(duplicity.backend.Backend):
    u"""Connect to remote store using File Transfer Protocol"""
    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        # we expect an output
        try:
            p = os.popen(u"lftp --version")
            fout = p.read()
            ret = p.close()
        except Exception:
            pass
        # there is no output if lftp not found
        if not fout:
            log.FatalError(u"LFTP not found:  Please install LFTP.",
                           log.ErrorCode.ftps_lftp_missing)

        # version is the second word of the second part of the first line
        version = fout.split(u'\n')[0].split(u' | ')[1].split()[1]
        log.Notice(u"LFTP version is %s" % version)

        self.parsed_url = parsed_url

        self.scheme = duplicity.backend.strip_prefix(parsed_url.scheme, u'lftp').lower()
        self.scheme = re.sub(u'^webdav', u'http', self.scheme)
        self.url_string = self.scheme + u'://' + parsed_url.hostname
        if parsed_url.port:
            self.url_string += u":%s" % parsed_url.port

        self.remote_path = re.sub(u'^/', u'', parsed_url.path)

        # Fix up an empty remote path
        if len(self.remote_path) == 0:
            self.remote_path = u'/'

        # Use an explicit directory name.
        if self.remote_path[-1] != u'/':
            self.remote_path += u'/'

        self.authflag = u''
        if self.parsed_url.username:
            self.username = self.parsed_url.username
            self.password = self.get_password()
            self.authflag = u"-u '%s,%s'" % (self.username, self.password)

        if config.ftp_connection == u'regular':
            self.conn_opt = u'off'
        else:
            self.conn_opt = u'on'

        # check for cacert file if https
        self.cacert_file = config.ssl_cacert_file
        if self.scheme == u'https' and not config.ssl_no_check_certificate:
            cacert_candidates = [u"~/.duplicity/cacert.pem",
                                 u"~/duplicity_cacert.pem",
                                 u"/etc/duplicity/cacert.pem"]
            # look for a default cacert file
            if not self.cacert_file:
                for path in cacert_candidates:
                    path = os.path.expanduser(path)
                    if (os.path.isfile(path)):
                        self.cacert_file = path
                        break

        # save config into a reusable temp file
        self.tempfd, self.tempname = tempdir.default().mkstemp()
        self.tempfile = os.fdopen(self.tempfd, u"w")
        self.tempfile.write(u"set ssl:verify-certificate " +
                            (u"false" if config.ssl_no_check_certificate else u"true") + u"\n")
        if self.cacert_file:
            self.tempfile.write(u"set ssl:ca-file " + cmd_quote(self.cacert_file) + u"\n")
        if config.ssl_cacert_path:
            self.tempfile.write(u"set ssl:ca-path " + cmd_quote(config.ssl_cacert_path) + u"\n")
        if self.parsed_url.scheme == u'ftps':
            self.tempfile.write(u"set ftp:ssl-allow true\n")
            self.tempfile.write(u"set ftp:ssl-protect-data true\n")
            self.tempfile.write(u"set ftp:ssl-protect-list true\n")
        elif self.parsed_url.scheme == u'ftpes':
            self.tempfile.write(u"set ftp:ssl-force on\n")
            self.tempfile.write(u"set ftp:ssl-protect-data on\n")
            self.tempfile.write(u"set ftp:ssl-protect-list on\n")
        else:
            self.tempfile.write(u"set ftp:ssl-allow false\n")
        self.tempfile.write(u"set http:use-propfind true\n")
        self.tempfile.write(u"set net:timeout %s\n" % config.timeout)
        self.tempfile.write(u"set net:max-retries %s\n" % config.num_retries)
        self.tempfile.write(u"set ftp:passive-mode %s\n" % self.conn_opt)
        if log.getverbosity() >= log.DEBUG:
            self.tempfile.write(u"debug\n")
        if self.parsed_url.scheme == u'ftpes':
            self.tempfile.write(u"open %s %s\n" % (self.authflag, self.url_string.replace(u'ftpes', u'ftp')))
        else:
            self.tempfile.write(u"open %s %s\n" % (self.authflag, self.url_string))
        self.tempfile.close()
        # print settings in debug mode
        if log.getverbosity() >= log.DEBUG:
            f = open(self.tempname, u'r')
            log.Debug(u"SETTINGS: \n"
                      u"%s" % f.read())

    def _put(self, source_path, remote_filename):
        if isinstance(remote_filename, b"".__class__):
            remote_filename = util.fsdecode(remote_filename)
        commandline = u"lftp -c \"source %s; mkdir -p %s; put %s -o %s\"" % (
            self.tempname,
            cmd_quote(self.remote_path),
            cmd_quote(source_path.uc_name),
            cmd_quote(self.remote_path) + util.fsdecode(remote_filename)
        )
        log.Debug(u"CMD: %s" % commandline)
        s, l, e = self.subprocess_popen(commandline)
        log.Debug(u"STATUS: %s" % s)
        log.Debug(u"STDERR:\n"
                  u"%s" % (e))
        log.Debug(u"STDOUT:\n"
                  u"%s" % (l))

    def _get(self, remote_filename, local_path):
        if isinstance(remote_filename, b"".__class__):
            remote_filename = util.fsdecode(remote_filename)
        commandline = u"lftp -c \"source %s; get %s -o %s\"" % (
            cmd_quote(self.tempname),
            cmd_quote(self.remote_path) + remote_filename,
            cmd_quote(local_path.uc_name)
        )
        log.Debug(u"CMD: %s" % commandline)
        _, l, e = self.subprocess_popen(commandline)
        log.Debug(u"STDERR:\n"
                  u"%s" % (e))
        log.Debug(u"STDOUT:\n"
                  u"%s" % (l))

    def _list(self):
        # Do a long listing to avoid connection reset
        # remote_dir = urllib.unquote(self.parsed_url.path.lstrip('/')).rstrip()
        remote_dir = urllib.parse.unquote(self.parsed_url.path)
        # print remote_dir
        quoted_path = cmd_quote(self.remote_path)
        # failing to cd into the folder might be because it was not created already
        commandline = u"lftp -c \"source %s; ( cd %s && ls ) || ( mkdir -p %s && cd %s && ls )\"" % (
            cmd_quote(self.tempname),
            quoted_path, quoted_path, quoted_path
        )
        log.Debug(u"CMD: %s" % commandline)
        _, l, e = self.subprocess_popen(commandline)
        log.Debug(u"STDERR:\n"
                  u"%s" % (e))
        log.Debug(u"STDOUT:\n"
                  u"%s" % (l))

        # Look for our files as the last element of a long list line
        return [util.fsencode(x.split()[-1]) for x in l.split(u'\n') if x]

    def _delete(self, filename):
        commandline = u"lftp -c \"source %s; cd %s; rm %s\"" % (
            cmd_quote(self.tempname),
            cmd_quote(self.remote_path),
            cmd_quote(util.fsdecode(filename))
        )
        log.Debug(u"CMD: %s" % commandline)
        _, l, e = self.subprocess_popen(commandline)
        log.Debug(u"STDERR:\n"
                  u"%s" % (e))
        log.Debug(u"STDOUT:\n"
                  u"%s" % (l))


duplicity.backend.register_backend(u"ftp", LFTPBackend)
duplicity.backend.register_backend(u"ftps", LFTPBackend)
duplicity.backend.register_backend(u"fish", LFTPBackend)
duplicity.backend.register_backend(u"ftpes", LFTPBackend)

duplicity.backend.register_backend(u"lftp+ftp", LFTPBackend)
duplicity.backend.register_backend(u"lftp+ftps", LFTPBackend)
duplicity.backend.register_backend(u"lftp+fish", LFTPBackend)
duplicity.backend.register_backend(u"lftp+ftpes", LFTPBackend)
duplicity.backend.register_backend(u"lftp+sftp", LFTPBackend)
duplicity.backend.register_backend(u"lftp+webdav", LFTPBackend)
duplicity.backend.register_backend(u"lftp+webdavs", LFTPBackend)
duplicity.backend.register_backend(u"lftp+http", LFTPBackend)
duplicity.backend.register_backend(u"lftp+https", LFTPBackend)

duplicity.backend.uses_netloc.extend([u'ftp', u'ftps', u'fish', u'ftpes',
                                      u'lftp+ftp', u'lftp+ftps',
                                      u'lftp+fish', u'lftp+ftpes',
                                      u'lftp+sftp',
                                      u'lftp+webdav', u'lftp+webdavs',
                                      u'lftp+http', u'lftp+https']
                                     )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

import config
import sys, unittest
sys.path.insert(0, u"../")

from duplicity import diffdir
from duplicity import patchdir
from duplicity import selection
from duplicity.path import *  # pylint: disable=unused-wildcard-import,redefined-builtin

config.setup()

class RootTest(unittest.TestCase):
    u"""Test doing operations that only root can"""

    def setUp(self):
        # must run with euid/egid of root
        assert(os.geteuid() == 0)
        # make sure uid/gid match euid/egid
        os.setuid(os.geteuid())
        os.setgid(os.getegid())
        assert not os.system(u"tar xzf manual/rootfiles.tar.gz > /dev/null 2>&1")

    def tearDown(self):
        assert not os.system(u"rm -rf /tmp/testfiles tempdir temp2.tar")

    def copyfileobj(self, infp, outfp):
        u"""Copy in fileobj to out, closing afterwards"""
        blocksize = 32 * 1024
        while 1:
            buf = infp.read(blocksize)
            if not buf: break
            outfp.write(buf)
        assert not infp.close()
        assert not outfp.close()

    def deltmp(self):
        u"""Delete temporary directories"""
        assert not os.system(u"rm -rf /tmp/testfiles/output")
        os.mkdir(u"/tmp/testfiles/output")

    def get_sel(self, path):
        u"""Get selection iter over the given directory"""
        return selection.Select(path).set_iter()

    def total_sequence(self, filelist):
        u"""Test signatures, diffing, and patching on directory list"""
        assert len(filelist) >= 2
        self.deltmp()
        assert not os.system(u"cp -pR %s /tmp/testfiles/output/sequence" %
                             (filelist[0],))
        seq_path = Path(u"/tmp/testfiles/output/sequence")
        sig = Path(u"/tmp/testfiles/output/sig.tar")
        diff = Path(u"/tmp/testfiles/output/diff.tar")
        for dirname in filelist[1:]:
            new_path = Path(dirname)
            diffdir.write_block_iter(
                diffdir.DirSig(selection.Select(seq_path).set_iter()), sig)

            diffdir.write_block_iter(
                diffdir.DirDelta(selection.Select(new_path).set_iter(),
                                 sig.open(u"rb")),
                diff)

            patchdir.Patch(seq_path, diff.open(u"rb"))

            assert seq_path.compare_recursive(new_path, 1)

    def test_basic_cycle(self):
        u"""Test cycle on dir with devices, changing uid/gid, etc."""
        self.total_sequence([u'/tmp/testfiles/root1', u'/tmp/testfiles/root2'])

    def test_patchdir(self):
        u"""Test changing uid/gid, devices"""
        self.deltmp()
        os.system(u"cp -pR /tmp/testfiles/root1 /tmp/testfiles/output/sequence")
        seq_path = Path(u"/tmp/testfiles/output/sequence")
        new_path = Path(u"/tmp/testfiles/root2")
        sig = Path(u"/tmp/testfiles/output/sig.tar")
        diff = Path(u"/tmp/testfiles/output/diff.tar")

        diffdir.write_block_iter(diffdir.DirSig(self.get_sel(seq_path)), sig)
        deltablock = diffdir.DirDelta(self.get_sel(new_path), sig.open(u"rb"))
        diffdir.write_block_iter(deltablock, diff)

        patchdir.Patch(seq_path, diff.open(u"rb"))

        # since we are not running as root, don't even both comparing,
        # just make sure file5 exists and file4 doesn't.
        file5 = seq_path.append(u"file5")
        assert file5.isreg()
        file4 = seq_path.append(u"file4")
        assert file4.type is None

    def test_patchdir2(self):
        u"""Again test files we don't have access to, this time Tar_WriteSig"""
        self.deltmp()
        sig_path = Path(u"/tmp/testfiles/output/sig.sigtar")
        tar_path = Path(u"/tmp/testfiles/output/tar.tar")
        basis_path = Path(u"/tmp/testfiles/root1")

        deltablock = diffdir.DirFull_WriteSig(self.get_sel(basis_path),
                                              sig_path.open(u"wb"))
        diffdir.write_block_iter(deltablock, tar_path)

def runtests(): unittest.main()

if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import print_function
from future import standard_library
standard_library.install_aliases()

import unittest

from . import UnitTestCase
from duplicity import path
from duplicity.statistics import *  # pylint: disable=redefined-builtin, unused-wildcard-import
from testing import _runtest_dir


class StatsObjTest(UnitTestCase):
    u"""Test StatsObj class"""
    def setUp(self):
        super(StatsObjTest, self).setUp()
        self.unpack_testfiles()

    def set_obj(self, s):
        u"""Set values of s's statistics"""
        s.SourceFiles = 1
        s.SourceFileSize = 2
        s.NewFiles = 3
        s.NewFileSize = 4
        s.DeletedFiles = 5
        s.ChangedFiles = 7
        s.ChangedFileSize = 8
        s.ChangedDeltaSize = 9
        s.DeltaEntries = 10
        s.RawDeltaSize = 11
        s.TotalDestinationSizeChange = 12
        s.StartTime = 13
        s.EndTime = 14

    def test_get_stats(self):
        u"""Test reading and writing stat objects"""
        s = StatsObj()
        assert s.get_stat(u'SourceFiles') is None
        self.set_obj(s)
        assert s.get_stat(u'SourceFiles') == 1

        s1 = StatsDeltaProcess()
        assert s1.get_stat(u'SourceFiles') == 0

    def test_get_stats_string(self):
        u"""Test conversion of stat object into string"""
        s = StatsObj()
        stats_string = s.get_stats_string()
        assert stats_string == u"", stats_string

        self.set_obj(s)
        stats_string = s.get_stats_string()
        assert stats_string == u"""\
StartTime 13.00 (Wed Dec 31 18:00:13 1969)
EndTime 14.00 (Wed Dec 31 18:00:14 1969)
ElapsedTime 1.00 (1 second)
SourceFiles 1
SourceFileSize 2 (2 bytes)
NewFiles 3
NewFileSize 4 (4 bytes)
DeletedFiles 5
ChangedFiles 7
ChangedFileSize 8 (8 bytes)
ChangedDeltaSize 9 (9 bytes)
DeltaEntries 10
RawDeltaSize 11 (11 bytes)
TotalDestinationSizeChange 12 (12 bytes)
""", u"'%s'" % stats_string

    def test_line_string(self):
        u"""Test conversion to a single line"""
        s = StatsObj()
        self.set_obj(s)
        statline = s.get_stats_line((u"sample", u"index", u"w", u"new\nline"))
        assert statline == u"sample/index/w/new\\nline 1 2 3 4 5 7 8 9 10 11", \
            repr(statline)

        statline = s.get_stats_line(())
        assert statline == u". 1 2 3 4 5 7 8 9 10 11"

        statline = s.get_stats_line((u"file name with spaces",))
        assert statline == (u"file\\x20name\\x20with\\x20spaces "
                            u"1 2 3 4 5 7 8 9 10 11"), repr(statline)

    def test_byte_summary(self):
        u"""Test conversion of bytes to strings like 7.23MB"""
        s = StatsObj()
        f = s.get_byte_summary_string
        assert f(1) == u"1 byte"
        assert f(234.34) == u"234 bytes"
        assert f(2048) == u"2.00 KB"
        assert f(3502243) == u"3.34 MB"
        assert f(314992230) == u"300 MB"
        assert f(36874871216) == u"34.3 GB", f(36874871216)
        assert f(3775986812573450) == u"3434 TB"

    def test_init_stats(self):
        u"""Test setting stat object from string"""
        s = StatsObj()
        s.set_stats_from_string(u"NewFiles 3 hello there")
        for attr in s.stat_attrs:
            if attr == u'NewFiles':
                assert s.get_stat(attr) == 3
            else:
                assert s.get_stat(attr) is None, (attr, s.__dict__[attr])

        s1 = StatsObj()
        self.set_obj(s1)
        assert not s1.stats_equal(s)

        s2 = StatsObj()
        s2.set_stats_from_string(s1.get_stats_string())
        assert s1.stats_equal(s2)

    def test_write_path(self):
        u"""Test reading and writing of statistics object"""
        p = path.Path(u"{0}/testfiles/statstest".format(_runtest_dir))
        if p.exists():
            p.delete()
        s = StatsObj()
        self.set_obj(s)
        s.write_stats_to_path(p)

        s2 = StatsObj()
        assert not s2.stats_equal(s)
        s2.read_stats_from_path(p)
        assert s2.stats_equal(s)

    def testAverage(self):
        u"""Test making an average statsobj"""
        s1 = StatsObj()
        s1.StartTime = 5
        s1.EndTime = 10
        s1.ElapsedTime = 5
        s1.ChangedFiles = 2
        s1.SourceFiles = 100
        s1.NewFileSize = 4

        s2 = StatsObj()
        s2.StartTime = 25
        s2.EndTime = 35
        s2.ElapsedTime = 10
        s2.ChangedFiles = 1
        s2.SourceFiles = 50
        s2.DeletedFiles = 0

        s3 = StatsObj().set_to_average([s1, s2])
        assert s3.StartTime is s3.EndTime is None
        assert s3.ElapsedTime == 7.5
        assert s3.DeletedFiles is s3.NewFileSize is None, (s3.DeletedFiles,
                                                           s3.NewFileSize)
        assert s3.ChangedFiles == 1.5
        assert s3.SourceFiles == 75


if __name__ == u"__main__":
    unittest.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4; encoding:utf8 -*-
#
# Copyright 2002 Ben Escoto <ben@emerose.org>
# Copyright 2007 Kenneth Loafman <kenneth@loafman.com>
#
# This file is part of duplicity.
#
# Duplicity is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Duplicity is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with duplicity; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

from __future__ import division
from builtins import str
from concurrent.futures import ThreadPoolExecutor
import os
import time

import duplicity.backend
from duplicity import config
from duplicity import log
from duplicity.errors import FatalBackendException, BackendException
from duplicity import progress
from duplicity import util

BOTO_MIN_VERSION = u"2.1.1"


def get_connection(scheme, parsed_url, storage_uri):
    try:
        from boto.s3.connection import S3Connection
        assert hasattr(S3Connection, u'lookup')

        # Newer versions of boto default to using
        # virtual hosting for buckets as a result of
        # upstream deprecation of the old-style access
        # method by Amazon S3. This change is not
        # backwards compatible (in particular with
        # respect to upper case characters in bucket
        # names); so we default to forcing use of the
        # old-style method unless the user has
        # explicitly asked us to use new-style bucket
        # access.
        #
        # Note that if the user wants to use new-style
        # buckets, we use the subdomain calling form
        # rather than given the option of both
        # subdomain and vhost. The reason being that
        # anything addressable as a vhost, is also
        # addressable as a subdomain. Seeing as the
        # latter is mostly a convenience method of
        # allowing browse:able content semi-invisibly
        # being hosted on S3, the former format makes
        # a lot more sense for us to use - being
        # explicit about what is happening (the fact
        # that we are talking to S3 servers).

        try:
            from boto.s3.connection import OrdinaryCallingFormat
            from boto.s3.connection import SubdomainCallingFormat
            cfs_supported = True
            calling_format = OrdinaryCallingFormat()
        except ImportError:
            cfs_supported = False
            calling_format = None

        if config.s3_use_new_style:
            if cfs_supported:
                calling_format = SubdomainCallingFormat()
            else:
                log.FatalError(u"Use of new-style (subdomain) S3 bucket addressing was"
                               u"requested, but does not seem to be supported by the "
                               u"boto library. Either you need to upgrade your boto "
                               u"library or duplicity has failed to correctly detect "
                               u"the appropriate support.",
                               log.ErrorCode.boto_old_style)
        else:
            if cfs_supported:
                calling_format = OrdinaryCallingFormat()
            else:
                calling_format = None

    except ImportError:
        log.FatalError(u"This backend (s3) requires boto library, version %s or later, "
                       u"(http://code.google.com/p/boto/)." % BOTO_MIN_VERSION,
                       log.ErrorCode.boto_lib_too_old)

    if not parsed_url.hostname:
        # Use the default host.
        conn = storage_uri.connect(is_secure=(not config.s3_unencrypted_connection))
    else:
        assert scheme == u's3'
        conn = storage_uri.connect(host=parsed_url.hostname, port=parsed_url.port,
                                   is_secure=(not config.s3_unencrypted_connection))

    if hasattr(conn, u'calling_format'):
        if calling_format is None:
            log.FatalError(u"It seems we previously failed to detect support for calling "
                           u"formats in the boto library, yet the support is there. This is "
                           u"almost certainly a duplicity bug.",
                           log.ErrorCode.boto_calling_format)
        else:
            conn.calling_format = calling_format

    else:
        # Duplicity hangs if boto gets a null bucket name.
        # HC: Caught a socket error, trying to recover
        raise BackendException(u'Boto requires a bucket name.')
    return conn


class BotoBackend(duplicity.backend.Backend):
    u"""
    Backend for Amazon's Simple Storage System, (aka Amazon S3), though
    the use of the boto module, (http://code.google.com/p/boto/).

    To make use of this backend you must set aws_access_key_id
    and aws_secret_access_key in your ~/.boto or /etc/boto.cfg
    with your Amazon Web Services key id and secret respectively.
    Alternatively you can export the environment variables
    AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY.
    """

    def __init__(self, parsed_url):
        duplicity.backend.Backend.__init__(self, parsed_url)

        try:
            import boto  # pylint: disable=import-error
            from boto.s3.connection import Location
        except ImportError:
            raise

        assert boto.Version >= BOTO_MIN_VERSION

        # This folds the null prefix and all null parts, which means that:
        #  //MyBucket/ and //MyBucket are equivalent.
        #  //MyBucket//My///My/Prefix/ and //MyBucket/My/Prefix are equivalent.
        self.url_parts = [x for x in parsed_url.path.split(u'/') if x != u'']

        if self.url_parts:
            self.bucket_name = self.url_parts.pop(0)
        else:
            # Duplicity hangs if boto gets a null bucket name.
            # HC: Caught a socket error, trying to recover
            raise BackendException(u'Boto requires a bucket name.')

        self.scheme = parsed_url.scheme

        if self.url_parts:
            self.key_prefix = u'%s/' % u'/'.join(self.url_parts)
        else:
            self.key_prefix = u''

        self.straight_url = duplicity.backend.strip_auth_from_url(parsed_url)
        self.parsed_url = parsed_url

        # duplicity and boto.storage_uri() have different URI formats.
        # boto uses scheme://bucket[/name] and specifies hostname on connect()
        self.boto_uri_str = u'://'.join((parsed_url.scheme[:2],
                                         parsed_url.path.lstrip(u'/')))
        if config.s3_european_buckets:
            self.my_location = Location.EU
        else:
            self.my_location = u''
        self.resetConnection()
        self._listed_keys = {}

    def _close(self):
        del self._listed_keys
        self._listed_keys = {}
        self.bucket = None
        self.conn = None
        self.storage_uri = None
        del self.conn
        del self.storage_uri

    def resetConnection(self):
        import boto  # pylint: disable=import-error

        if getattr(self, u'conn', False):
            self.conn.close()
        self.bucket = None
        self.conn = None
        self.storage_uri = None
        del self.conn
        del self.storage_uri
        self.storage_uri = boto.storage_uri(self.boto_uri_str)
        self.conn = get_connection(self.scheme, self.parsed_url, self.storage_uri)
        if not self.conn.lookup(self.bucket_name):
            self.bucket = self.conn.create_bucket(self.bucket_name,
                                                  location=self.my_location)
        else:
            self.bucket = self.conn.get_bucket(self.bucket_name)

    def _retry_cleanup(self):
        self.resetConnection()

    def _put(self, source_path, remote_filename):
        remote_filename = util.fsdecode(remote_filename)

        if config.s3_european_buckets:
            if not config.s3_use_new_style:
                raise FatalBackendException(u"European bucket creation was requested, but not new-style "
                                            u"bucket addressing (--s3-use-new-style)",
                                            code=log.ErrorCode.s3_bucket_not_style)

        if self.bucket is None:
            try:
                self.bucket = self.conn.get_bucket(self.bucket_name)
            except Exception as e:
                if u"NoSuchBucket" in str(e):
                    self.bucket = self.conn.create_bucket(self.bucket_name,
                                                          location=self.my_location)
                else:
                    raise

        key = self.bucket.new_key(self.key_prefix + remote_filename)

        if config.s3_use_rrs:
            storage_class = u'REDUCED_REDUNDANCY'
        elif config.s3_use_ia:
            storage_class = u'STANDARD_IA'
        elif config.s3_use_onezone_ia:
            storage_class = u'ONEZONE_IA'
        elif config.s3_use_glacier and u"manifest" not in remote_filename:
            storage_class = u'GLACIER'
        elif config.s3_use_glacier_ir and u"manifest" not in remote_filename:
            storage_class = u'GLACIER_IR'
        else:
            storage_class = u'STANDARD'
        log.Info(u"Uploading %s/%s to %s Storage" % (self.straight_url, remote_filename, storage_class))
        if config.s3_use_sse:
            headers = {
                u'Content-Type': u'application/octet-stream',
                u'x-amz-storage-class': storage_class,
                u'x-amz-server-side-encryption': u'AES256'
            }
        elif config.s3_use_sse_kms:
            if config.s3_kms_key_id is None:
                raise FatalBackendException(u"S3 USE SSE KMS was requested, but key id not provided "
                                            u"require (--s3-kms-key-id)",
                                            code=log.ErrorCode.s3_kms_no_id)
            headers = {
                u'Content-Type': u'application/octet-stream',
                u'x-amz-storage-class': storage_class,
                u'x-amz-server-side-encryption': u'aws:kms',
                u'x-amz-server-side-encryption-aws-kms-key-id': config.s3_kms_key_id
            }
            if config.s3_kms_grant is not None:
                headers[u'x-amz-grant-full-control'] = config.s3_kms_grant
        else:
            headers = {
                u'Content-Type': u'application/octet-stream',
                u'x-amz-storage-class': storage_class
            }

        upload_start = time.time()
        self.upload(source_path.name, key, headers)
        upload_end = time.time()
        total_s = abs(upload_end - upload_start) or 1  # prevent a zero value!
        rough_upload_speed = os.path.getsize(source_path.name) / total_s
        log.Debug(u"Uploaded %s/%s to %s Storage at roughly %f bytes/second" %
                  (self.straight_url, remote_filename, storage_class,
                   rough_upload_speed))

    def _get(self, remote_filename, local_path):
        remote_filename = util.fsdecode(remote_filename)
        key_name = self.key_prefix + remote_filename
        self.pre_process_download(remote_filename, wait=True)
        key = self._listed_keys[key_name]
        self.resetConnection()
        key.get_contents_to_filename(local_path.name)

    def _list(self):
        if not self.bucket:
            raise BackendException(u"No connection to backend")
        return self.list_filenames_in_bucket()

    def list_filenames_in_bucket(self):
        # We add a 'd' to the prefix to make sure it is not null (for boto) and
        # to optimize the listing of our filenames, which always begin with 'd'.
        # This will cause a failure in the regression tests as below:
        #   FAIL: Test basic backend operations
        #   <tracback snipped>
        #   AssertionError: Got list: []
        #   Wanted: ['testfile']
        # Because of the need for this optimization, it should be left as is.
        # for k in self.bucket.list(prefix = self.key_prefix + 'd', delimiter = '/'):
        filename_list = []
        for k in self.bucket.list(prefix=self.key_prefix):
            try:
                filename = k.key.replace(self.key_prefix, u'', 1)
                filename_list.append(filename)
                self._listed_keys[k.key] = k
                log.Debug(u"Listed %s/%s" % (self.straight_url, filename))
            except AttributeError:
                pass
        return filename_list

    def _delete(self, filename):
        filename = util.fsdecode(filename)
        self.bucket.delete_key(self.key_prefix + filename)

    def _query(self, filename):
        filename = util.fsdecode(filename)
        key = self.bucket.lookup(self.key_prefix + filename)
        if key is None:
            return {u'size': -1}
        return {u'size': key.size}

    def upload(self, filename, key, headers):
        key.set_contents_from_filename(filename, headers,
                                       cb=progress.report_transfer,
                                       num_cb=(max(2, 8 * config.volsize / (1024 * 1024)))
                                       )  # Max num of callbacks = 8 times x megabyte
        key.close()

    def pre_process_download(self, remote_filename, wait=False):
        # Used primarily to move files in Glacier to S3
        key_name = self.key_prefix + remote_filename
        if not self._listed_keys.get(key_name, False):
            self._listed_keys[key_name] = list(self.bucket.list(key_name))[0]
        key = self._listed_keys[key_name]

        if key.storage_class == u"GLACIER":
            # We need to move the file out of glacier
            if not self.bucket.get_key(key.key).ongoing_restore:
                log.Info(u"File %s is in Glacier storage, restoring to S3" % remote_filename)
                key.restore(days=1)  # Shouldn't need this again after 1 day
            if wait:
                log.Info(u"Waiting for file %s to restore from Glacier" % remote_filename)
                while self.bucket.get_key(key.key).ongoing_restore:
                    time.sleep(60)
                    self.resetConnection()
                log.Info(u"File %s was successfully restored from Glacier" % remote_filename)

    def pre_process_download_batch(self, remote_filenames):
        log.Info(u"Starting batch unfreezing from Glacier")
        # Used primarily to move all necessary files in Glacier to S3 at once
        with ThreadPoolExecutor(thread_name_prefix=u's3-unfreeze-glacier') as executor:
            for remote_filename in remote_filenames:
                remote_filename = util.fsdecode(remote_filename)
                executor.submit(self.pre_process_download, remote_filename, False)
        log.Info(u"Batch unfreezing from Glacier finished")
